try:
    from mpire.dashboard.dashboard import connect_to_dashboard, start_dashboard
except (ImportError, ModuleNotFoundError):
    def connect_to_dashboard(*_, **__):
        raise NotImplementedError("Install the dashboard dependencies to enable the dashboard")

    def start_dashboard(*_, **__):
        raise NotImplementedError("Install the dashboard dependencies to enable the dashboard")



import heapq
import itertools
import math
import os
from datetime import datetime, timedelta
from multiprocessing import Array, cpu_count
from typing import Callable, Generator, Iterable, List, Optional, Tuple, Union

try:
    import numpy as np
    NUMPY_INSTALLED = True
except ImportError:
    np = None
    NUMPY_INSTALLED = False

from mpire.context import RUNNING_WINDOWS

# Needed for setting CPU affinity
if RUNNING_WINDOWS:
    try:
        import win32api
        import win32con
        import win32process
        WIN32API_AVAILABLE = True
        WIN32API_ERROR = None
    except ImportError as e:
        WIN32API_AVAILABLE = False
        WIN32API_ERROR = e
        WIN32API_ERROR.msg += " If you're using Conda, you can run `conda install pywin32` to install the missing " \
                              "module."


def set_cpu_affinity(pid: int, mask: List[int]) -> None:
    """
    Sets the CPU affinity for a given process.

    On Windows-based systems with more than 64 processors, I'm not sure if this will work. See
    https://docs.microsoft.com/en-us/windows/win32/api/winbase/nf-winbase-setprocessaffinitymask#parameters.

    :param pid: Process ID
    :param mask: List of CPU IDs
    """
    if RUNNING_WINDOWS:
        # On Conda systems simply install pywin32 doesn't always work. In those cases, you need to run
        # `conda install pywin32`.
        if not WIN32API_AVAILABLE:
            raise WIN32API_ERROR

        # Convert mask to something Windows understands
        windows_mask = 0
        for cpu_id in mask:
            windows_mask ^= 2 ** cpu_id

        # Get handle and set affinity
        handle = win32api.OpenProcess(win32con.PROCESS_ALL_ACCESS, True, pid)
        win32process.SetProcessAffinityMask(handle, windows_mask)
    else:
        os.sched_setaffinity(pid, mask)


def chunk_tasks(iterable_of_args: Iterable, iterable_len: Optional[int] = None,
                chunk_size: Optional[Union[int, float]] = None, n_splits: Optional[int] = None) \
        -> Generator[Iterable, None, None]:
    """
    Chunks tasks such that individual workers will receive chunks of tasks rather than individual ones, which can
    speed up processing drastically.

    :param iterable_of_args: A numpy array or an iterable containing tuples of arguments to pass to a worker, which
        passes it to the function
    :param iterable_len: Number of tasks available in ``iterable_of_args``. Only needed when ``iterable_of_args`` is a
        generator
    :param chunk_size: Number of simultaneous tasks to give to a worker. If ``None``, will use ``n_splits`` to determine
        the chunk size
    :param n_splits: Number of splits to use when ``chunk_size`` is ``None``
    :return: Generator of chunked task arguments
    """
    if chunk_size is None and n_splits is None:
        raise ValueError("chunk_size and n_splits cannot both be None")

    # Determine chunk size
    if chunk_size is None:
        # Get number of tasks
        if iterable_len is not None:
            n_tasks = iterable_len
        elif hasattr(iterable_of_args, '__len__'):
            n_tasks = len(iterable_of_args)
        else:
            raise ValueError('Failed to obtain length of iterable when chunk size is None. Remedy: either provide an '
                             'iterable with a len() function or specify iterable_len in the function call')

        # Determine chunk size
        chunk_size = n_tasks / n_splits

    # Chunk tasks
    args_iter = iter(iterable_of_args)
    current_chunk_size = chunk_size
    n_elements_returned = 0
    while True:
        # Use numpy slicing if available. We use max(1, ...) to always at least get one element
        if NUMPY_INSTALLED and isinstance(iterable_of_args, np.ndarray):
            chunk = iterable_of_args[n_elements_returned:n_elements_returned + max(1, math.ceil(current_chunk_size))]
        else:
            chunk = tuple(itertools.islice(args_iter, max(1, math.ceil(current_chunk_size))))

        # If we ran out of input, we stop
        if len(chunk) == 0:
            return

        # If the iterable has more elements than the given iterable length, we stop
        if iterable_len is not None and n_elements_returned + len(chunk) > iterable_len:
            chunk = chunk[:iterable_len - n_elements_returned]
            if chunk:
                yield chunk
            return

        yield chunk
        current_chunk_size = (current_chunk_size + chunk_size) - math.ceil(current_chunk_size)
        n_elements_returned += len(chunk)


def apply_numpy_chunking(iterable_of_args: Iterable, iterable_len: Optional[int] = None,
                         chunk_size: Optional[int] = None, n_splits: Optional[int] = None,
                         n_jobs: Optional[int] = None) -> Tuple[Iterable, int, int, None]:
    """
    If we're dealing with numpy arrays, chunk them using numpy slicing and return changed map parameters

    :param iterable_of_args: A numpy array or an iterable containing tuples of arguments to pass to a worker, which
        passes it to the function
    :param iterable_len: When chunk_size is set to ``None`` it needs to know the number of tasks. This can either be
        provided by implementing the ``__len__`` function on the iterable object, or by specifying the number of tasks
    :param chunk_size: Number of simultaneous tasks to give to a worker. If ``None``, will generate ``n_jobs * 4``
        number of chunks
    :param n_splits: Number of splits to use when ``chunk_size`` is ``None``
    :param n_jobs: Number of workers to spawn. If ``None``, will use ``cpu_count()``.
    :return: Chunked ``iterable_of_args`` with updated ``iterable_len``, ``chunk_size`` and ``n_splits``
    """
    if iterable_len is not None:
        iterable_of_args = iterable_of_args[:iterable_len]
    iterable_len = get_n_chunks(iterable_of_args, iterable_len, chunk_size, n_splits, n_jobs)
    iterable_of_args = make_single_arguments(chunk_tasks(iterable_of_args, len(iterable_of_args), chunk_size,
                                                         n_splits or (n_jobs * 4 if n_jobs is not None else None)))
    chunk_size = 1
    n_splits = None

    return iterable_of_args, iterable_len, chunk_size, n_splits


def get_n_chunks(iterable_of_args: Iterable, iterable_len: Optional[int] = None, chunk_size: Optional[int] = None,
                 n_splits: Optional[int] = None, n_jobs: Optional[int] = None) -> int:
    """
    Get number of chunks

    :param iterable_of_args: A numpy array or an iterable containing tuples of arguments to pass to a worker, which
        passes it to the function
    :param iterable_len: Number of tasks available in ``iterable_of_args``. Only needed when ``iterable_of_args`` is a
        generator
    :param chunk_size: Number of simultaneous tasks to give to a worker. If ``None``, will use ``n_splits`` to determine
        the chunk size
    :param n_splits: Number of splits to use when ``chunk_size`` is ``None``
    :param n_jobs: Number of workers to spawn. If ``None``, will use ``cpu_count()``
    :return: Number of chunks that will be created by the chunker
    """
    # Get number of tasks
    if iterable_len is not None:
        n_tasks = min(iterable_len, len(iterable_of_args)) if hasattr(iterable_of_args, '__len__') else iterable_len
    elif hasattr(iterable_of_args, '__len__'):
        n_tasks = len(iterable_of_args)
    else:
        raise ValueError('Failed to obtain length of iterable. Remedy: either provide an iterable with a len() '
                         'function or specify iterable_len in the function call')

    # Determine chunk size
    if chunk_size is None:
        chunk_size = n_tasks / (n_splits or (n_jobs or cpu_count()) * 4)

    return min(n_tasks, math.ceil(n_tasks / chunk_size))


def make_single_arguments(iterable_of_args: Iterable, generator: bool = True) -> Union[List, Generator]:
    """
    Converts an iterable of single arguments to an iterable of single argument tuples

    :param iterable_of_args: A numpy array or an iterable containing tuples of arguments to pass to a worker, which
        passes it to the function
    :param generator: Whether or not to return a generator, otherwise a materialized list will be returned
    :return: Iterable of single argument tuples
    """
    gen = ((arg,) for arg in iterable_of_args)
    return gen if generator else list(gen)


def format_seconds(seconds: Optional[Union[int, float]], with_milliseconds: bool) -> str:
    """
    Format seconds to a string, without milliseconds

    :param seconds: Number of seconds
    :param with_milliseconds: Whether to display milliseconds as well
    :return: String formatted time
    """
    if seconds is None:
        return ''

    # Format to hours:minutes:seconds.milliseconds. Only the first 3 digits of the milliseconds is shown
    duration = str(timedelta(seconds=seconds)).rsplit('.', 1)
    if with_milliseconds and len(duration) > 1:
        duration = f'{duration[0]}.{duration[1][:3]}'
    else:
        duration = duration[0]

    return duration


class TimeIt:

    """ Simple class that provides a context manager for keeping track of task duration and adds the total number
     of seconds in a designated output array """

    def __init__(self, cum_time_array: Optional[Array], array_idx: int, max_time_array: Optional[Array] = None,
                 format_args_func: Optional[Callable] = None) -> None:
        """
        :param cum_time_array: Optional array to store cumulative time in
        :param array_idx: Index of cum_time_array to store the time value to
        :param max_time_array: Optional array to store maximum time duration in. Note that the array_idx doesn't apply
            to this array. The entire array is used for heapq
        :param format_args_func: Optional function which should return the formatted args corresponding to the function
            called within this context manager
        """
        self.cum_time_array = cum_time_array
        self.array_idx = array_idx
        self.max_time_array = max_time_array
        self.format_args_func = format_args_func
        self.start_dt = None

    def __enter__(self) -> None:
        self.start_dt = datetime.now()

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        duration = (datetime.now() - self.start_dt).total_seconds()
        if self.cum_time_array is not None:
            self.cum_time_array[self.array_idx] += duration
        if self.max_time_array is not None and duration > self.max_time_array[0][0]:
            heapq.heappushpop(self.max_time_array,
                              (duration, self.format_args_func() if self.format_args_func is not None else None))



import itertools
import multiprocessing as mp
import warnings
from typing import Any, Callable, Iterable, List, Optional, Sized, Tuple, Union

from tqdm import tqdm

from mpire.context import RUNNING_WINDOWS

# Typedefs
CPUList = Optional[List[Union[int, List[int]]]]


class WorkerPoolParams:

    """
    Data class, with some parameter verification, for all :obj:`mpire.WorkerPool` parameters.
    """

    def __init__(self, n_jobs: Optional[int] = None, daemon: bool = True, cpu_ids: CPUList = None,
                 shared_objects: Any = None, pass_worker_id: bool = False, use_worker_state: bool = False,
                 start_method: str = 'fork', keep_alive: bool = False, use_dill: bool = False) -> None:
        """
        See ``WorkerPool.__init__`` docstring.
        """
        self.n_jobs = n_jobs or mp.cpu_count()
        self.daemon = daemon
        self.cpu_ids = self._check_cpu_ids(cpu_ids)
        self.shared_objects = shared_objects
        self.pass_worker_id = pass_worker_id
        self.use_worker_state = use_worker_state
        self.start_method = start_method
        self.keep_alive = keep_alive
        self.use_dill = use_dill

        # Number of (chunks of) jobs a child process can process before requesting a restart
        self.worker_lifespan = None

        # User provided functions to call
        self.func = None
        self.worker_init = None
        self.worker_exit = None
        self.enable_insights = None

    def _check_cpu_ids(self, cpu_ids: CPUList) -> List[List[int]]:
        """
        Checks the cpu_ids parameter for correctness

        :param cpu_ids: List of CPU IDs to use for pinning child processes to specific CPUs. The list must be as long as
            the number of jobs used (if ``n_jobs`` equals ``None`` it must be equal to ``mpire.cpu_count()``), or the
            list must have exactly one element. In the former case, element x specifies the CPU ID(s) to use for child
            process x. In the latter case the single element specifies the CPU ID(s) for all child  processes to use. A
            single element can be either a single integer specifying a single CPU ID, or a list of integers specifying
            that a single child process can make use of multiple CPU IDs. If ``None``, CPU pinning will be disabled.
            Note that CPU pinning may only work on Linux based systems
        :return: cpu_ids
        """
        # Check CPU IDs
        converted_cpu_ids = []
        if cpu_ids:
            # Check number of arguments
            if len(cpu_ids) != 1 and len(cpu_ids) != self.n_jobs:
                raise ValueError("Number of CPU IDs (%d) does not match number of jobs (%d)" %
                                 (len(cpu_ids), self.n_jobs))

            # Convert CPU IDs to proper format and find the max and min CPU ID
            max_cpu_id = 0
            min_cpu_id = 0
            for cpu_id in cpu_ids:
                if isinstance(cpu_id, list):
                    converted_cpu_ids.append(cpu_id)
                    max_cpu_id = max(max_cpu_id, max(cpu for cpu in cpu_id))
                    min_cpu_id = min(min_cpu_id, min(cpu for cpu in cpu_id))
                elif isinstance(cpu_id, int):
                    converted_cpu_ids.append([cpu_id])
                    max_cpu_id = max(max_cpu_id, cpu_id)
                    min_cpu_id = min(min_cpu_id, cpu_id)
                else:
                    raise TypeError("CPU ID(s) must be either a list or a single integer")

            # Check max CPU ID
            if max_cpu_id >= mp.cpu_count():
                raise ValueError("CPU ID %d exceeds the maximum CPU ID available on your system: %d" %
                                 (max_cpu_id, mp.cpu_count() - 1))

            # Check min CPU ID
            if min_cpu_id < 0:
                raise ValueError("CPU IDs cannot be negative")

        # If only one item is given, use this item for all child processes
        if len(converted_cpu_ids) == 1:
            converted_cpu_ids = list(itertools.repeat(converted_cpu_ids[0], self.n_jobs))

        return converted_cpu_ids

    def check_map_parameters(self, iterable_of_args: Union[Sized, Iterable], iterable_len: Optional[int],
                             max_tasks_active: Optional[int], chunk_size: Optional[Union[int, float]],
                             n_splits: Optional[int], worker_lifespan: Optional[int], progress_bar: bool,
                             progress_bar_position: int) \
            -> Tuple[Optional[int], int, Optional[int], Union[bool, tqdm]]:
        """
        Check the parameters provided to any (i)map function. Also extracts the number of tasks and can modify the
        ``chunk_size`` and ``progress_bar`` parameters.

        :param iterable_of_args: A numpy array or an iterable containing tuples of arguments to pass to a worker, which
            passes it to the function
        :param iterable_len: Number of elements in the ``iterable_of_args``. When chunk_size is set to ``None`` it needs
            to know the number of tasks. This can either be provided by implementing the ``__len__`` function on the
            iterable object, or by specifying the number of tasks
        :param max_tasks_active: Maximum number of active tasks in the queue. Use ``None`` to not limit the queue
        :param chunk_size: Number of simultaneous tasks to give to a worker. If ``None`` it will generate ``n_jobs * 4``
            number of chunks
        :param n_splits: Number of splits to use when ``chunk_size`` is ``None``
        :param worker_lifespan: Number of chunks a worker can handle before it is restarted. If ``None``, workers will
            stay alive the entire time. Use this when workers use up too much memory over the course of time
        :param progress_bar: When ``True`` it will display a progress bar
        :param progress_bar_position: Denotes the position (line nr) of the progress bar. This is useful wel using
            multiple progress bars at the same time
        :return: Number of tasks, max tasks active, chunk size, progress bar
        """
        # Get number of tasks
        n_tasks = None
        if iterable_len is not None:
            n_tasks = iterable_len
        elif hasattr(iterable_of_args, '__len__'):
            n_tasks = len(iterable_of_args)
        elif chunk_size is None or progress_bar:
            warnings.warn('Failed to obtain length of iterable when chunk size or number of splits is None and/or a '
                          'progress bar is requested. Chunk size is set to 1 and no progress bar will be shown. '
                          'Remedy: either provide an iterable with a len() function or specify iterable_len in the '
                          'function call', RuntimeWarning, stacklevel=2)
            chunk_size = 1
            progress_bar = False

        # Check chunk_size parameter
        if chunk_size is not None:
            if not isinstance(chunk_size, (int, float)):
                raise TypeError('chunk_size should be either None or an integer value')
            elif chunk_size <= 0:
                raise ValueError('chunk_size should be a positive integer > 0')

        # Check n_splits parameter (only when chunk_size is None)
        else:
            if not (isinstance(n_splits, int) or n_splits is None):
                raise TypeError('n_splits should be either None or an integer value')
            if isinstance(n_splits, int) and n_splits <= 0:
                raise ValueError('n_splits should be a positive integer > 0')

        # Check max_tasks_active parameter
        if max_tasks_active is None:
            max_tasks_active = self.n_jobs * 2
        elif isinstance(max_tasks_active, int):
            if max_tasks_active <= 0:
                raise ValueError('max_tasks_active should be a positive integer or None')
        else:
            raise TypeError('max_tasks_active should be a positive integer or None')

        # If worker lifespan is not None or not a positive integer, raise
        if isinstance(worker_lifespan, int):
            if worker_lifespan <= 0:
                raise ValueError('worker_lifespan should be either None or a positive integer (> 0)')
        elif worker_lifespan is not None:
            raise TypeError('worker_lifespan should be either None or a positive integer (> 0)')

        # Progress bar is currently not supported on Windows when using threading as start method. For reasons still
        # unknown we get a TypeError: cannot pickle '_thread.Lock' object.
        if RUNNING_WINDOWS and progress_bar and self.start_method == "threading":
            raise ValueError("Progress bar is currently not supported on Windows when using start_method='threading'")

        # Progress bar position should be a positive integer
        if not isinstance(progress_bar_position, int):
            raise TypeError('progress_bar_position should be a positive integer (>= 0)')
        if progress_bar_position < 0:
            raise ValueError('progress_bar_position should be a positive integer (>= 0)')

        return n_tasks, max_tasks_active, chunk_size, progress_bar

    def set_map_params(self, func: Callable, worker_init: Optional[Callable], worker_exit: Optional[Callable],
                       worker_lifespan: int, enable_insights: bool) -> None:
        """
        Set map specific parameters

        :param func: Function to call each time new task arguments become available
        :param worker_init: Function to call each time a new worker starts
        :param worker_exit: Function to call each time a worker exits
        :param worker_lifespan: Number of chunks a worker can handle before it is restarted
        :param enable_insights: Whether to enable worker insights
        """
        self.func = func
        self.worker_init = worker_init
        self.worker_exit = worker_exit
        self.worker_lifespan = worker_lifespan
        self.enable_insights = enable_insights

    def workers_need_restart(self, func: Callable, worker_init: Optional[Callable], worker_exit: Optional[Callable],
                             worker_lifespan: int, enable_insights: bool) -> bool:
        """
        Checks if workers need to be restarted based on some key parameters

        :param func: Function to call each time new task arguments become available
        :param worker_init: Function to call each time a new worker starts
        :param worker_exit: Function to call each time a worker exits
        :param worker_lifespan: Number of chunks a worker can handle before it is restarted
        :param enable_insights: Whether to enable worker insights
        :return: Whether the workers need to be restarted
        """
        return (func != self.func or
                worker_init != self.worker_init or
                worker_exit != self.worker_exit or
                worker_lifespan != self.worker_lifespan or
                enable_insights != self.enable_insights)



import multiprocessing as mp
import os
import signal
import unittest

from mpire.context import RUNNING_WINDOWS
from mpire.signal import DelayedKeyboardInterrupt, DisableKeyboardInterruptSignal
from tests.utils import ConditionalDecorator


@ConditionalDecorator(unittest.skip("Signals aren't fully supported on Windows"), RUNNING_WINDOWS)
class DelayedKeyboardInterruptTest(unittest.TestCase):

    def test_delayed_keyboard_interrupt(self):
        """
        The process should delay the keyboard interrupt in case ``in_thread=False``, so the expected value should be 1.
        However, we can't send signals to threads and so the DelayedKeyboardInterrupt doesn't do anything in that case.
        So there's no point in testing this with threading
        """
        # Create events so we know when the process has started and we can send an interrupt
        started_event = mp.Event()
        quit_event = mp.Event()
        value = mp.Value('i', 0)

        # Start process and wait until it starts
        p = mp.Process(target=self.delayed_process_job, args=(started_event, quit_event, value))
        p.start()
        started_event.wait()

        # Send kill signal and wait for it to join
        os.kill(p.pid, signal.SIGINT)
        quit_event.set()
        p.join()

        # Verify expected value.
        self.assertEqual(value.value, 1)

    @staticmethod
    def delayed_process_job(started_event: mp.Event, quit_event: mp.Event, value: mp.Value):
        """
        Should be affected by interrupt
        """
        try:
            with DelayedKeyboardInterrupt(in_thread=False):
                started_event.set()
                quit_event.wait()
                value.value = 1
        except KeyboardInterrupt:
            pass
        else:
            value.value = 2


@ConditionalDecorator(unittest.skip("Signals aren't fully supported on Windows"), RUNNING_WINDOWS)
class DisabledKeyboardInterruptTest(unittest.TestCase):

    def test_disabled_keyboard_interrupt(self):
        """
        The process should ignore a keyboard interrupt entirely, which means the expected value should be True
        """
        # Create events so we know when the process has started and we can send an interrupt
        started_event = mp.Event()
        quit_event = mp.Event()
        value = mp.Value('b', False)
        p = mp.Process(target=self.disabled_process_job, args=(started_event, quit_event, value))
        p.start()
        started_event.wait()
        os.kill(p.pid, signal.SIGINT)
        quit_event.set()
        p.join()

        # If everything worked the value should be set to True
        self.assertEqual(value.value, True)

    @staticmethod
    def disabled_process_job(started_event: mp.Event, quit_event: mp.Event, value: mp.Value):
        """
        Should not be affected by interrupt
        """
        with DisableKeyboardInterruptSignal():
            started_event.set()
            quit_event.wait()
        value.value = True



import logging
import multiprocessing as mp
import os
import queue
import signal
import threading
from datetime import datetime
from typing import Any, Callable, Dict, Generator, Iterable, List, Optional, Sized, Union

try:
    import numpy as np
    NUMPY_INSTALLED = True
except ImportError:
    np = None
    NUMPY_INSTALLED = False

from mpire.comms import WorkerComms
from mpire.context import DEFAULT_START_METHOD, RUNNING_WINDOWS
from mpire.dashboard.connection_utils import get_dashboard_connection_details
from mpire.insights import WorkerInsights
from mpire.params import CPUList, WorkerPoolParams
from mpire.progress_bar import ProgressBarHandler
from mpire.signal import DisableKeyboardInterruptSignal
from mpire.tqdm_utils import TqdmManager
from mpire.utils import apply_numpy_chunking, chunk_tasks, set_cpu_affinity
from mpire.worker import MP_CONTEXTS, worker_factory

logger = logging.getLogger(__name__)


class WorkerPool:
    """
    A multiprocessing worker pool which acts like a ``multiprocessing.Pool``, but is faster and has more options.
    """

    def __init__(self, n_jobs: Optional[int] = None, daemon: bool = True, cpu_ids: CPUList = None,
                 shared_objects: Any = None, pass_worker_id: bool = False, use_worker_state: bool = False,
                 start_method: str = DEFAULT_START_METHOD, keep_alive: bool = False, use_dill: bool = False) -> None:
        """
        :param n_jobs: Number of workers to spawn. If ``None``, will use ``cpu_count()``
        :param daemon: Whether to start the child processes as daemon
        :param cpu_ids: List of CPU IDs to use for pinning child processes to specific CPUs. The list must be as long as
            the number of jobs used (if ``n_jobs`` equals ``None`` it must be equal to ``mpire.cpu_count()``), or the
            list must have exactly one element. In the former case, element x specifies the CPU ID(s) to use for child
            process x. In the latter case the single element specifies the CPU ID(s) for all child  processes to use. A
            single element can be either a single integer specifying a single CPU ID, or a list of integers specifying
            that a single child process can make use of multiple CPU IDs. If ``None``, CPU pinning will be disabled.
            Note that CPU pinning may only work on Linux based systems
        :param shared_objects: ``None`` or any other type of object (multiple objects can be wrapped in a single tuple).
            Shared objects is only passed on to the user function when it's not ``None``
        :param pass_worker_id: Whether to pass on a worker ID to the user function or not
        :param use_worker_state: Whether to let a worker have a worker state or not
        :param start_method: What process start method to use. Options for multiprocessing: ``'fork'`` (default, if
            available), ``'forkserver'`` and ``'spawn'`` (default, if ``'fork'`` isn't available). For multithreading
            use ``'threading'``. See https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods
            for more information and
            https://docs.python.org/3/library/multiprocessing.html#the-spawn-and-forkserver-start-methods for some
            caveats when using the ``'spawn'`` or ``'forkserver'`` methods
        :param keep_alive: When True it will keep workers alive after completing a map call, allowing to reuse workers
            when map is called with the same function and worker lifespan multiple times in a row
        :param use_dill: Whether to use dill as serialization backend. Some exotic types (e.g., lambdas, nested
            functions) don't work well when using ``spawn`` as start method. In such cased, use ``dill`` (can be a bit
            slower sometimes)
        """
        # Set parameters
        self.params = WorkerPoolParams(n_jobs, daemon, cpu_ids, shared_objects, pass_worker_id, use_worker_state,
                                       start_method, keep_alive, use_dill)

        # Worker factory
        self.Worker = worker_factory(start_method, use_dill)

        # Multiprocessing context
        if start_method == 'threading':
            self.ctx = MP_CONTEXTS['threading']
        else:
            self.ctx = MP_CONTEXTS['mp_dill' if use_dill else 'mp'][start_method]

        # Container of the child processes and corresponding communication objects
        self._workers = []
        self._worker_comms = WorkerComms(self.ctx, self.params.n_jobs, use_dill, start_method == 'threading')
        self._exit_results = None

        # Worker insights, used for profiling
        self._worker_insights = WorkerInsights(self.ctx, self.params.n_jobs)

    def pass_on_worker_id(self, pass_on: bool = True) -> None:
        """
        Set whether to pass on the worker ID to the function to be executed or not (default= ``False``).

        The worker ID will be the first argument passed on to the function

        :param pass_on: Whether to pass on a worker ID to the user function or not
        """
        self.params.pass_worker_id = pass_on

    def set_shared_objects(self, shared_objects: Any = None) -> None:
        """
        Set shared objects to pass to the workers.

        Shared objects will be copy-on-write. Process-aware shared objects (e.g., ``multiprocessing.Array``) can be used
        to write to the same object from multiple processes. When providing shared objects the provided function in the
        ``map`` function should receive the shared objects as its first argument if the worker ID is not passed on. If
        the worker ID is passed on the shared objects will be the second argument.

        :param shared_objects: ``None`` or any other type of object (multiple objects can be wrapped in a single tuple).
            Shared objects is only passed on to the user function when it's not ``None``
        """
        self.params.shared_objects = shared_objects

    def set_use_worker_state(self, use_worker_state: bool = True) -> None:
        """
        Set whether or not each worker should have its own state variable. Each worker has its own state, so it's not
        shared between the workers.

        :param use_worker_state: Whether to let a worker have a worker state or not
        """
        self.params.use_worker_state = use_worker_state

    def set_keep_alive(self, keep_alive: bool = True) -> None:
        """
        Set whether workers should be kept alive in between consecutive map calls.

        :param keep_alive: When True it will keep workers alive after completing a map call, allowing to reuse workers
            when map is called with the same function and worker lifespan multiple times in a row
        """
        self.params.keep_alive = keep_alive

    def _start_workers(self, progress_bar: bool, enable_insights: bool) -> None:
        """
        Spawns the workers and starts them so they're ready to start reading from the tasks queue.

        :param progress_bar: Whether there's a progress bar
        :param enable_insights: Whether to enable worker insights
        """
        # Init communication primitives
        logger.debug("Initializing comms")
        self._worker_comms.init_comms(self.params.worker_exit is not None, progress_bar)
        self._worker_insights.reset_insights(enable_insights)
        self._exit_results = []

        # Start new workers
        logger.debug("Spinning up workers")
        for worker_id in range(self.params.n_jobs):
            self._workers.append(self._start_worker(worker_id))
        logger.debug("Workers created")

    def _restart_workers(self) -> None:
        """
        Restarts workers that need to be restarted.
        """
        for worker_id in self._worker_comms.get_worker_restarts():
            # Obtain results from exit results queue (should be done before joining the worker)
            if self.params.worker_exit:
                self._exit_results.append(self._worker_comms.get_exit_results(worker_id))

            # Join worker
            self._worker_comms.reset_worker_restart(worker_id)
            self._workers[worker_id].join()

            # Start new worker
            self._workers[worker_id] = self._start_worker(worker_id)

    def _start_worker(self, worker_id: int) -> mp.Process:
        """
        Creates and starts a single worker

        :param worker_id: ID of the worker
        :return: Worker instance
        """
        # Disable the interrupt signal. We let the process die gracefully if it needs to
        with DisableKeyboardInterruptSignal():
            # Create worker
            w = self.Worker(worker_id, self.params, self._worker_comms, self._worker_insights,
                            TqdmManager.get_connection_details(), get_dashboard_connection_details(), datetime.now())
            w.daemon = self.params.daemon
            w.start()

            # Pin CPU if desired
            if self.params.cpu_ids:
                set_cpu_affinity(w.pid, self.params.cpu_ids[worker_id])

            return w

    def get_exit_results(self) -> List:
        """
        Obtain a list of exit results when an exit function is defined.

        :return: Exit results list
        """
        return self._exit_results

    def __enter__(self) -> 'WorkerPool':
        """
        Enable the use of the ``with`` statement.
        """
        return self

    def __exit__(self, *_: Any) -> None:
        """
        Enable the use of the ``with`` statement. Gracefully terminates workers, if there are any
        """
        if self._workers:
            self.terminate()

    def map(self, func: Callable, iterable_of_args: Union[Sized, Iterable], iterable_len: Optional[int] = None,
            max_tasks_active: Optional[int] = None, chunk_size: Optional[int] = None, n_splits: Optional[int] = None,
            worker_lifespan: Optional[int] = None, progress_bar: bool = False, progress_bar_position: int = 0,
            concatenate_numpy_output: bool = True, enable_insights: bool = False,
            worker_init: Optional[Callable] = None, worker_exit: Optional[Callable] = None) -> Any:
        """
        Same as ``multiprocessing.map()``. Also allows a user to set the maximum number of tasks available in the queue.
        Note that this function can be slower than the unordered version.

        :param func: Function to call each time new task arguments become available. When passing on the worker ID the
            function should receive the worker ID as its first argument. If shared objects are provided the function
            should receive those as the next argument. If the worker state has been enabled it should receive a state
            variable as the next argument
        :param iterable_of_args: A numpy array or an iterable containing tuples of arguments to pass to a worker, which
            passes it to the function ``func``
        :param iterable_len: Number of elements in the ``iterable_of_args``. When chunk_size is set to ``None`` it needs
            to know the number of tasks. This can either be provided by implementing the ``__len__`` function on the
            iterable object, or by specifying the number of tasks
        :param max_tasks_active: Maximum number of active tasks in the queue. If ``None`` it will be converted to
            ``n_jobs * 2``
        :param chunk_size: Number of simultaneous tasks to give to a worker. When ``None`` it will use ``n_splits``.
        :param n_splits: Number of splits to use when ``chunk_size`` is ``None``. When both ``chunk_size`` and
            ``n_splits`` are ``None``, it will use ``n_splits = n_jobs * 64``.
        :param worker_lifespan: Number of tasks a worker can handle before it is restarted. If ``None``, workers will
            stay alive the entire time. Use this when workers use up too much memory over the course of time
        :param progress_bar: When ``True`` it will display a progress bar
        :param progress_bar_position: Denotes the position (line nr) of the progress bar. This is useful wel using
            multiple progress bars at the same time
        :param concatenate_numpy_output: When ``True`` it will concatenate numpy output to a single numpy array
        :param enable_insights: Whether to enable worker insights. Might come at a small performance penalty (often
            neglible)
        :param worker_init: Function to call each time a new worker starts. When passing on the worker ID the function
            should receive the worker ID as its first argument. If shared objects are provided the function should
            receive those as the next argument. If the worker state has been enabled it should receive a state variable
            as the next argument
        :param worker_exit: Function to call each time a worker exits. Return values will be fetched and made available
            through :obj:`mpire.WorkerPool.get_exit_results`. When passing on the worker ID the function should receive
            the worker ID as its first argument. If shared objects are provided the function should receive those as the
            next argument. If the worker state has been enabled it should receive a state variable as the next argument
        :return: List with ordered results
        """
        # Notify workers to keep order in mind
        self._worker_comms.set_keep_order()

        # If we're dealing with numpy arrays, we have to chunk them here already
        if NUMPY_INSTALLED and isinstance(iterable_of_args, np.ndarray):
            iterable_of_args, iterable_len, chunk_size, n_splits = apply_numpy_chunking(iterable_of_args, iterable_len,
                                                                                        chunk_size, n_splits,
                                                                                        self.params.n_jobs)

        # Process all args
        if iterable_len is None and hasattr(iterable_of_args, '__len__'):
            iterable_len = len(iterable_of_args)
        results = self.map_unordered(func, ((args_idx, args) for args_idx, args in enumerate(iterable_of_args)),
                                     iterable_len, max_tasks_active, chunk_size, n_splits, worker_lifespan,
                                     progress_bar, progress_bar_position, enable_insights, worker_init, worker_exit)

        # Notify workers to forget about order
        self._worker_comms.clear_keep_order()

        # Rearrange and return
        sorted_results = [result[1] for result in sorted(results, key=lambda result: result[0])]

        # Convert back to numpy if necessary
        return (np.concatenate(sorted_results) if NUMPY_INSTALLED and sorted_results and concatenate_numpy_output and
                                                  isinstance(sorted_results[0], np.ndarray) else sorted_results)

    def map_unordered(self, func: Callable, iterable_of_args: Union[Sized, Iterable],
                      iterable_len: Optional[int] = None, max_tasks_active: Optional[int] = None,
                      chunk_size: Optional[int] = None, n_splits: Optional[int] = None,
                      worker_lifespan: Optional[int] = None, progress_bar: bool = False,
                      progress_bar_position: int = 0, enable_insights: bool = False,
                      worker_init: Optional[Callable] = None, worker_exit: Optional[Callable] = None) -> Any:
        """
        Same as ``multiprocessing.map()``, but unordered. Also allows a user to set the maximum number of tasks
        available in the queue.

        :param func: Function to call each time new task arguments become available. When passing on the worker ID the
            function should receive the worker ID as its first argument. If shared objects are provided the function
            should receive those as the next argument. If the worker state has been enabled it should receive a state
            variable as the next argument
        :param iterable_of_args: A numpy array or an iterable containing tuples of arguments to pass to a worker, which
            passes it to the function ``func``
        :param iterable_len: Number of elements in the ``iterable_of_args``. When chunk_size is set to ``None`` it needs
            to know the number of tasks. This can either be provided by implementing the ``__len__`` function on the
            iterable object, or by specifying the number of tasks
        :param max_tasks_active: Maximum number of active tasks in the queue. If ``None`` it will be converted to
            ``n_jobs * 2``
        :param chunk_size: Number of simultaneous tasks to give to a worker. When ``None`` it will use ``n_splits``.
        :param n_splits: Number of splits to use when ``chunk_size`` is ``None``. When both ``chunk_size`` and
            ``n_splits`` are ``None``, it will use ``n_splits = n_jobs * 64``.
        :param worker_lifespan: Number of tasks a worker can handle before it is restarted. If ``None``, workers will
            stay alive the entire time. Use this when workers use up too much memory over the course of time
        :param progress_bar: When ``True`` it will display a progress bar
        :param progress_bar_position: Denotes the position (line nr) of the progress bar. This is useful wel using
            multiple progress bars at the same time
        :param enable_insights: Whether to enable worker insights. Might come at a small performance penalty (often
            neglible)
        :param worker_init: Function to call each time a new worker starts. When passing on the worker ID the function
            should receive the worker ID as its first argument. If shared objects are provided the function should
            receive those as the next argument. If the worker state has been enabled it should receive a state variable
            as the next argument
        :param worker_exit: Function to call each time a worker exits. Return values will be fetched and made available
            through :obj:`mpire.WorkerPool.get_exit_results`. When passing on the worker ID the function should receive
            the worker ID as its first argument. If shared objects are provided the function should receive those as the
            next argument. If the worker state has been enabled it should receive a state variable as the next argument
        :return: List with unordered results
        """
        # Simply call imap and cast it to a list. This make sure all elements are there before returning
        return list(self.imap_unordered(func, iterable_of_args, iterable_len, max_tasks_active, chunk_size,
                                        n_splits, worker_lifespan, progress_bar, progress_bar_position,
                                        enable_insights, worker_init, worker_exit))

    def imap(self, func: Callable, iterable_of_args: Union[Sized, Iterable], iterable_len: Optional[int] = None,
             max_tasks_active: Optional[int] = None, chunk_size: Optional[int] = None, n_splits: Optional[int] = None,
             worker_lifespan: Optional[int] = None, progress_bar: bool = False,
             progress_bar_position: int = 0, enable_insights: bool = False, worker_init: Optional[Callable] = None,
             worker_exit: Optional[Callable] = None) -> Generator[Any, None, None]:
        """
        Same as ``multiprocessing.imap_unordered()``, but ordered. Also allows a user to set the maximum number of
        tasks available in the queue.

        :param func: Function to call each time new task arguments become available. When passing on the worker ID the
            function should receive the worker ID as its first argument. If shared objects are provided the function
            should receive those as the next argument. If the worker state has been enabled it should receive a state
            variable as the next argument
        :param iterable_of_args: A numpy array or an iterable containing tuples of arguments to pass to a worker, which
            passes it to the function ``func``
        :param iterable_len: Number of elements in the ``iterable_of_args``. When chunk_size is set to ``None`` it needs
            to know the number of tasks. This can either be provided by implementing the ``__len__`` function on the
            iterable object, or by specifying the number of tasks
        :param max_tasks_active: Maximum number of active tasks in the queue. If ``None`` it will be converted to
            ``n_jobs * 2``
        :param chunk_size: Number of simultaneous tasks to give to a worker. When ``None`` it will use ``n_splits``.
        :param n_splits: Number of splits to use when ``chunk_size`` is ``None``. When both ``chunk_size`` and
            ``n_splits`` are ``None``, it will use ``n_splits = n_jobs * 64``.
        :param worker_lifespan: Number of tasks a worker can handle before it is restarted. If ``None``, workers will
            stay alive the entire time. Use this when workers use up too much memory over the course of time
        :param progress_bar: When ``True`` it will display a progress bar
        :param progress_bar_position: Denotes the position (line nr) of the progress bar. This is useful wel using
            multiple progress bars at the same time
        :param enable_insights: Whether to enable worker insights. Might come at a small performance penalty (often
            neglible)
        :param worker_init: Function to call each time a new worker starts. When passing on the worker ID the function
            should receive the worker ID as its first argument. If shared objects are provided the function should
            receive those as the next argument. If the worker state has been enabled it should receive a state variable
            as the next argument
        :param worker_exit: Function to call each time a worker exits. Return values will be fetched and made available
            through :obj:`mpire.WorkerPool.get_exit_results`. When passing on the worker ID the function should receive
            the worker ID as its first argument. If shared objects are provided the function should receive those as the
            next argument. If the worker state has been enabled it should receive a state variable as the next argument
        :return: Generator yielding ordered results
        """
        # Notify workers to keep order in mind
        self._worker_comms.set_keep_order()

        # If we're dealing with numpy arrays, we have to chunk them here already
        if NUMPY_INSTALLED and isinstance(iterable_of_args, np.ndarray):
            iterable_of_args, iterable_len, chunk_size, n_splits = apply_numpy_chunking(iterable_of_args, iterable_len,
                                                                                        chunk_size, n_splits,
                                                                                        self.params.n_jobs)

        # Yield results in order
        next_result_idx = 0
        tmp_results = {}
        if iterable_len is None and hasattr(iterable_of_args, '__len__'):
            iterable_len = len(iterable_of_args)
        for result_idx, result in self.imap_unordered(func, ((args_idx, args) for args_idx, args
                                                             in enumerate(iterable_of_args)), iterable_len,
                                                      max_tasks_active, chunk_size, n_splits, worker_lifespan,
                                                      progress_bar, progress_bar_position, enable_insights, worker_init,
                                                      worker_exit):

            # Check if the next one(s) to return is/are temporarily stored. We use a while-true block with dict.pop() to
            # keep the temporary store as small as possible
            while True:
                if next_result_idx in tmp_results:
                    yield tmp_results.pop(next_result_idx)
                    next_result_idx += 1
                else:
                    break

            # Check if the current result is the next one to return. If so, return it
            if result_idx == next_result_idx:
                yield result
                next_result_idx += 1
            # Otherwise, temporarily store the current result
            else:
                tmp_results[result_idx] = result

        # Yield all remaining results
        for result_idx in sorted(tmp_results.keys()):
            yield tmp_results.pop(result_idx)

        # Notify workers to forget about order
        self._worker_comms.clear_keep_order()

    def imap_unordered(self, func: Callable, iterable_of_args: Union[Sized, Iterable],
                       iterable_len: Optional[int] = None, max_tasks_active: Optional[int] = None,
                       chunk_size: Optional[int] = None, n_splits: Optional[int] = None,
                       worker_lifespan: Optional[int] = None, progress_bar: bool = False,
                       progress_bar_position: int = 0, enable_insights: bool = False,
                       worker_init: Optional[Callable] = None,
                       worker_exit: Optional[Callable] = None) -> Generator[Any, None, None]:
        """
        Same as ``multiprocessing.imap_unordered()``. Also allows a user to set the maximum number of tasks available in
        the queue.

        :param func: Function to call each time new task arguments become available. When passing on the worker ID the
            function should receive the worker ID as its first argument. If shared objects are provided the function
            should receive those as the next argument. If the worker state has been enabled it should receive a state
            variable as the next argument
        :param iterable_of_args: A numpy array or an iterable containing tuples of arguments to pass to a worker, which
            passes it to the function ``func``
        :param iterable_len: Number of elements in the ``iterable_of_args``. When chunk_size is set to ``None`` it needs
            to know the number of tasks. This can either be provided by implementing the ``__len__`` function on the
            iterable object, or by specifying the number of tasks
        :param max_tasks_active: Maximum number of active tasks in the queue. If ``None`` it will be converted to
            ``n_jobs * 2``
        :param chunk_size: Number of simultaneous tasks to give to a worker. When ``None`` it will use ``n_splits``.
        :param n_splits: Number of splits to use when ``chunk_size`` is ``None``. When both ``chunk_size`` and
            ``n_splits`` are ``None``, it will use ``n_splits = n_jobs * 64``.
        :param worker_lifespan: Number of tasks a worker can handle before it is restarted. If ``None``, workers will
            stay alive the entire time. Use this when workers use up too much memory over the course of time
        :param progress_bar: When ``True`` it will display a progress bar
        :param progress_bar_position: Denotes the position (line nr) of the progress bar. This is useful wel using
            multiple progress bars at the same time
        :param enable_insights: Whether to enable worker insights. Might come at a small performance penalty (often
            neglible)
        :param worker_init: Function to call each time a new worker starts. When passing on the worker ID the function
            should receive the worker ID as its first argument. If shared objects are provided the function should
            receive those as the next argument. If the worker state has been enabled it should receive a state variable
            as the next argument
        :param worker_exit: Function to call each time a worker exits. Return values will be fetched and made available
            through :obj:`mpire.WorkerPool.get_exit_results`. When passing on the worker ID the function should receive
            the worker ID as its first argument. If shared objects are provided the function should receive those as the
            next argument. If the worker state has been enabled it should receive a state variable as the next argument
        :return: Generator yielding unordered results
        """
        # If we're dealing with numpy arrays, we have to chunk them here already
        iterator_of_chunked_args = []
        numpy_chunking = False
        if NUMPY_INSTALLED and isinstance(iterable_of_args, np.ndarray):
            iterator_of_chunked_args, iterable_len, chunk_size, n_splits = apply_numpy_chunking(
                iterable_of_args, iterable_len, chunk_size, n_splits, self.params.n_jobs
            )
            numpy_chunking = True

        # Check parameters and thereby obtain the number of tasks. The chunk_size and progress bar parameters could be
        # modified as well
        n_tasks, max_tasks_active, chunk_size, progress_bar = self.params.check_map_parameters(
            iterable_of_args, iterable_len, max_tasks_active, chunk_size, n_splits, worker_lifespan, progress_bar,
            progress_bar_position
        )

        # Start tqdm manager if a progress bar is desired. Will only start one when not already started. This has to be
        # done before starting the workers in case nested pools are used
        tqdm_manager_owner = TqdmManager.start_manager() if progress_bar else False

        # Chunk the function arguments. Make single arguments when we're not dealing with numpy arrays
        if not numpy_chunking:
            iterator_of_chunked_args = chunk_tasks(iterable_of_args, n_tasks, chunk_size,
                                                   n_splits or self.params.n_jobs * 64)

        # Start workers if there aren't any. If they already exist they must be restarted when either the function(s) to
        # execute, worker_lifespan, or enable_insights changes
        if self._workers and self.params.workers_need_restart(func, worker_init, worker_exit, worker_lifespan,
                                                              enable_insights):
            self.stop_and_join(None, keep_alive=False)
        if not self._workers:
            self.params.set_map_params(func, worker_init, worker_exit, worker_lifespan, enable_insights)
            self._start_workers(progress_bar, enable_insights)

        # Create progress bar handler, which receives progress updates from the workers and updates the progress bar
        # accordingly
        try:
            with ProgressBarHandler(self.ctx, self.params.start_method == 'threading', func, self.params.n_jobs,
                                    progress_bar, n_tasks, progress_bar_position, self._worker_comms,
                                    self._worker_insights) as progress_bar_handler:
                try:
                    # Process all args in the iterable
                    n_active = 0
                    while not self._worker_comms.exception_thrown():
                        # Add task, only if allowed and if there are any
                        if n_active < max_tasks_active:
                            try:
                                self._worker_comms.add_task(next(iterator_of_chunked_args))
                                n_active += 1
                            except StopIteration:
                                break

                        # Check if new results are available, but don't wait for it
                        try:
                            yield from self._worker_comms.get_results(block=False)
                            n_active -= 1
                        except queue.Empty:
                            pass

                        # Restart workers if necessary
                        self._restart_workers()

                    # Obtain the results not yet obtained
                    while not self._worker_comms.exception_thrown() and n_active != 0:
                        try:
                            yield from self._worker_comms.get_results(block=True, timeout=0.01)
                            n_active -= 1
                        except queue.Empty:
                            pass

                        # Restart workers if necessary
                        self._restart_workers()

                    # Terminate if exception has been thrown at this point
                    if self._worker_comms.exception_thrown():
                        self._handle_exception(progress_bar_handler)

                    # All results are in: it's clean up time
                    self.stop_and_join(progress_bar_handler, keep_alive=self.params.keep_alive)

                except KeyboardInterrupt:
                    self._handle_exception(progress_bar_handler)

            # Join exception queue, if it hasn't already
            logger.debug("Joining exception queue")
            self._worker_comms.join_exception_queue(self.params.keep_alive)
            logger.debug("Done joining exception queue, now joining progress bar queue")
            self._worker_comms.join_progress_bar_task_completed_queue(self.params.keep_alive)
            logger.debug("Done joining progress bar queue")

        finally:
            if tqdm_manager_owner:
                TqdmManager.stop_manager()

        # Log insights
        if enable_insights:
            logger.debug(self._worker_insights.get_insights_string())

    def _handle_exception(self, progress_bar_handler: Optional[ProgressBarHandler] = None) -> None:
        """
        Handles exceptions thrown by workers and KeyboardInterrupts

        :param progress_bar_handler: Progress bar handler
        """
        logger.debug("Exception occurred")

        # If we have a progress bar and this function is called because there was a KeyboardInterrupt (i.e., no
        # exception thrown), we pass the KeyboardInterrupt to the progress bar handler (using the exception queue)
        keyboard_interrupt = False
        with self._worker_comms.exception_lock:
            if not self._worker_comms.exception_thrown():
                keyboard_interrupt = True
                self._worker_comms.set_exception_thrown()
                if self._worker_comms.has_progress_bar() and progress_bar_handler is not None:
                    self._worker_comms.add_exception(KeyboardInterrupt, "KeyboardInterrupt")

        # When we're not dealing with a KeyboardInterrupt, obtain the exception thrown by a worker
        if keyboard_interrupt:
            err, traceback_str = KeyboardInterrupt, ""
        else:
            err, traceback_str = self._worker_comms.get_exception()
            self._worker_comms.task_done_exception()
        logger.debug(f"Exception obtained: {err}")

        # Join exception queue and progress bar, and terminate workers
        logger.debug("Joining exception queue")
        self._worker_comms.join_exception_queue()
        logger.debug("Exception queue joined")
        if progress_bar_handler and progress_bar_handler.process is not None:
            logger.debug("Joining progress bar handler")
            progress_bar_handler.process.join()
            logger.debug("Progress bar handler joined")
        self.terminate()

        # Clear keep order event so we can safely reuse the WorkerPool and use (i)map_unordered after an (i)map call
        self._worker_comms.clear_keep_order()

        # Raise
        logger.debug("Re-raising obtained exception")
        raise err(traceback_str)

    def stop_and_join(self, progress_bar_handler: Optional[ProgressBarHandler] = None,
                      keep_alive: bool = False) -> None:
        """
        When ``keep_alive=False``: inserts a poison pill, grabs the exit results, waits until the tasks/results queues
        are done, and wait until all workers are finished.
        When ``keep_alive=True``: inserts a non-lethal poison pill, and waits until the tasks/results queues are done.

        Note that the results queue should be drained first before joining the workers, otherwise we can get a deadlock.
        For more information, see the warnings at:
        https://docs.python.org/3.4/library/multiprocessing.html#pipes-and-queues.

        :param progress_bar_handler: Progress bar handler
        :param keep_alive: Whether to keep the workers alive
        """
        if self._workers:
            logger.debug(f"Cleaning up workers (keep_alive={keep_alive})")

            # All tasks have been processed and results are in. Insert (non-lethal) poison pill
            if keep_alive:
                self._worker_comms.insert_non_lethal_poison_pill()
            else:
                self._worker_comms.insert_poison_pill()

            # Wait until all (non-lethal) poison pills have been consumed. When a worker's lifetime has been reached
            # just before consuming the poison pill, we need to restart them
            logger.debug("Joining task queues")
            t = threading.Thread(target=self._worker_comms.join_task_queues, args=(keep_alive,))
            t.daemon = True
            t.start()
            while not self._worker_comms.exception_thrown():
                t.join(timeout=0.01)
                if not t.is_alive():
                    break
                self._restart_workers()
            logger.debug("Done joining task queues")

            # When an exception occurred in the above process (i.e., the worker init function raises), we need to handle
            # the exception (i.e., terminate and raise)
            if self._worker_comms.exception_thrown():
                self._handle_exception(progress_bar_handler)

            # Obtain results from the exit results queues (should be done before joining the workers)
            if not keep_alive and self.params.worker_exit:
                logger.debug("Obtaining exit results")
                self._exit_results.extend(self._worker_comms.get_exit_results_all_workers())
                self._worker_comms.set_all_exit_results_obtained()
                logger.debug("Done obtaining exit results")

            # If an exception occurred in the exit function, we need to handle the exception (i.e., terminate and raise)
            if self._worker_comms.exception_thrown():
                self._handle_exception(progress_bar_handler)

            # Join (exit) results queue (should be done immediately, as all (exit) results have been processed)
            logger.debug("Joining results queues")
            self._worker_comms.join_results_queues(keep_alive)
            logger.debug("Done joining results queues")

            # Join workers
            if not keep_alive:
                logger.debug("Joining workers")
                for worker_process in self._workers:
                    worker_process.join()
                    # Added since Python 3.7
                    if hasattr(worker_process, 'close'):
                        worker_process.close()
                self._workers = []
                logger.debug("Done joining workers")

            logger.debug("Cleaning up workers done")

    def terminate(self) -> None:
        """
        Tries to do a graceful shutdown of the workers, by interrupting them. In the case processes deadlock it will
        send a sigkill.
        """
        logger.debug("Terminating workers")

        # Set exception thrown so workers know to stop fetching new tasks
        self._worker_comms.set_exception_thrown()

        # When we're working with threads we have to wait for them to join. We can't kill threads in Python
        if self.params.start_method == 'threading':
            threads = self._workers
        else:
            # Create cleanup threads such that processes can get killed simultaneously, which can save quite some time
            threads = []
            dont_wait_event = threading.Event()
            dont_wait_event.set()
            for worker_id, worker_process in enumerate(self._workers):
                t = threading.Thread(target=self._terminate_worker, args=(worker_id, worker_process, dont_wait_event))
                t.start()
                threads.append(t)

        # Wait until cleanup threads are done
        for t in threads:
            t.join()
        logger.debug("Terminating workers done")

        # Drain and join the queues
        logger.debug("Draining queues")
        self._worker_comms.drain_queues()
        logger.debug("Draining queues done")

        # Reset workers
        self._workers = []

    def _terminate_worker(self, worker_id: int, worker_process: mp.context.Process,
                          dont_wait_event: threading.Event) -> None:
        """
        Terminates a single worker process

        :param worker_id: Worker ID
        :param worker_process: Worker instance
        :param dont_wait_event: Event object to indicate whether other termination threads should continue. I.e., when
            we set it to False, threads should wait.
        """
        # We wait until workers are done terminating. However, we don't have all the patience in the world. When the
        # patience runs out we terminate them.
        try_count = 10
        while try_count > 0:
            # Signal handling in Windows is cumbersome, to say the least. Therefore, it handles error handling
            # differently. See Worker::_exit_gracefully_windows for more information.
            if not RUNNING_WINDOWS:
                try:
                    os.kill(worker_process.pid, signal.SIGUSR1)
                except ProcessLookupError:
                    pass

            self._worker_comms.wait_for_dead_worker(worker_id, timeout=0.1)
            if not self._worker_comms.is_worker_alive(worker_id):
                break

            # For properly joining, it can help if we try to get some results here. Workers can still be busy putting
            # items in queues under the hood
            self._worker_comms.drain_queues_terminate_worker(worker_id, dont_wait_event)
            try_count -= 1
            if not dont_wait_event.is_set():
                dont_wait_event.wait()

        # Join the worker process
        try_count = 10
        while try_count > 0:
            try:
                worker_process.join(timeout=0.1)
                if not worker_process.is_alive():
                    break
            except ValueError:
                # For Windows compatibility
                pass

            # For properly joining, it can help if we try to get some results here. Workers can still be busy putting
            # items in queues under the hood, even at this point.
            self._worker_comms.drain_queues_terminate_worker(worker_id, dont_wait_event)
            try_count -= 1
            if not dont_wait_event.is_set():
                dont_wait_event.wait()

        # If, after all this, the worker is still alive, we terminate it with a brutal kill signal. This shouldn't
        # really happen. But, better safe than sorry
        try:
            if worker_process.is_alive():
                worker_process.terminate()
                worker_process.join()
        except ValueError:
            # For Windows compatibility
            pass

        # Added since Python 3.7
        if hasattr(worker_process, 'close'):
            worker_process.close()

    def print_insights(self) -> None:
        """
        Prints insights per worker
        """
        print(self._worker_insights.get_insights_string())

    def get_insights(self) -> Dict:
        """
        Creates insights from the raw insight data

        :return: Dictionary containing worker insights
        """
        return self._worker_insights.get_insights()



import multiprocessing as mp
try:
    import multiprocess as mp_dill
except ImportError:
    mp_dill = None
import platform
import threading

# Check if fork is available as start method. It's not available on Windows machines
try:
    mp.get_context('fork')
    FORK_AVAILABLE = True
except ValueError:
    FORK_AVAILABLE = False

# Check if we're running on Windows
RUNNING_WINDOWS = platform.system() == "Windows"


# Threading context so we can use threading as backend as well
class ThreadingContext:

    Barrier = threading.Barrier
    Event = threading.Event
    Lock = threading.Lock
    Thread = threading.Thread

    # threading doesn't have Array and JoinableQueue, so we take it from multiprocessing. Both are thread-safe. We need
    # the Process class for the MPIRE insights SyncManager instance.
    Array = mp.Array
    JoinableQueue = mp.JoinableQueue
    Process = mp.Process
    Value = mp.Value


MP_CONTEXTS = {'mp': {'fork': mp.get_context('fork') if FORK_AVAILABLE else None,
                      'forkserver': mp.get_context('forkserver') if FORK_AVAILABLE else None,
                      'spawn': mp.get_context('spawn')},
               'threading': ThreadingContext}
if mp_dill is not None:
    MP_CONTEXTS['mp_dill'] = {'fork': mp_dill.get_context('fork') if FORK_AVAILABLE else None,
                              'forkserver': mp_dill.get_context('forkserver') if FORK_AVAILABLE else None,
                              'spawn': mp_dill.get_context('spawn')}

DEFAULT_START_METHOD = 'fork' if FORK_AVAILABLE else 'spawn'



import types
import unittest
from datetime import datetime
from itertools import chain, product
from multiprocessing import cpu_count
from unittest.mock import patch

import numpy as np

from mpire.utils import apply_numpy_chunking, chunk_tasks, format_seconds, get_n_chunks, make_single_arguments, TimeIt
from tests.utils import MockDatetimeNow


class ChunkTasksTest(unittest.TestCase):

    def test_no_chunk_size_no_n_splits_provided(self):
        """
        Test that a ValueError is raised when no chunk_size and n_splits are provided
        """
        with self.assertRaises(ValueError):
            next(chunk_tasks([]))

    def test_generator_without_iterable_len(self):
        """
        Test that a ValueError is raised when a generator is provided without iterable_len
        """
        with self.assertRaises(ValueError):
            next(chunk_tasks(iter([]), n_splits=1))

    def test_chunk_size_has_priority_over_n_splits(self):
        """
        Test that chunk_size is prioritized over n_splits
        """
        chunks = list(chunk_tasks(range(4), chunk_size=4, n_splits=4))
        self.assertEqual(len(chunks), 1)
        self.assertEqual(len(chunks[0]), 4)
        self.assertEqual(list(range(4)), list(chain.from_iterable(chunks)))

    def test_empty_input(self):
        """
        Test that the chunker is an empty generator for an empty input iterable
        """
        with self.subTest('list input'):
            chunks = list(chunk_tasks([], n_splits=5))
            self.assertEqual(len(chunks), 0)

        with self.subTest('generator/iterator input'):
            chunks = list(chunk_tasks(iter([]), iterable_len=0, n_splits=5))
            self.assertEqual(len(chunks), 0)

    def test_iterable_len_doesnt_match_input_size(self):
        """
        Test for cases where iterable_len does and does not match the number of arguments (it should work fine)
        """
        num_args = 10
        for iter_len in [5, 10, 20]:
            expected_args_sum = min(iter_len, num_args)

            # Test for normal list (range is considered a normal list as it implements __len__ and such)
            with self.subTest(iter_len=iter_len, input='list'):
                chunks = list(chunk_tasks(range(num_args), iterable_len=iter_len, n_splits=1))
                total_args = sum(map(len, chunks))
                self.assertEqual(total_args, expected_args_sum)
                self.assertEqual(list(range(expected_args_sum)), list(chain.from_iterable(chunks)))

            # Test for an actual generator (range does not really behave like one)
            with self.subTest(iter_len=iter_len, input='generator/iterator'):
                chunks = list(chunk_tasks(iter(range(num_args)), iterable_len=iter_len, n_splits=1))
                total_args = sum(map(len, chunks))
                self.assertEqual(total_args, expected_args_sum)
                self.assertEqual(list(range(expected_args_sum)), list(chain.from_iterable(chunks)))

    def test_n_splits(self):
        """
        Test different values of n_splits: len(args) {<, ==, >} n_splits
        """
        n_splits = 5
        for num_args in [n_splits - 1, n_splits, n_splits + 1]:
            expected_n_chunks = min(n_splits, num_args)

            # Test for normal list (range is considered a normal list as it implements __len__ and such)
            with self.subTest(num_args=num_args, input='list'):
                chunks = list(chunk_tasks(range(num_args), n_splits=n_splits))
                self.assertEqual(len(chunks), expected_n_chunks)
                self.assertEqual(list(range(num_args)), list(chain.from_iterable(chunks)))

            # Test for an actual generator (range does not really behave like one)
            with self.subTest(num_args=num_args, input='generator/iterator'):
                chunks = list(chunk_tasks(iter(range(num_args)), iterable_len=num_args, n_splits=n_splits))
                self.assertEqual(len(chunks), expected_n_chunks)
                self.assertEqual(list(range(num_args)), list(chain.from_iterable(chunks)))

    def test_chunk_size(self):
        """
        Test that chunks are of the right size if chunk_size is provided
        """
        chunk_size = 3
        for num_args in [chunk_size - 1, chunk_size, chunk_size + 1]:

            # Test for normal list (range is considered a normal list as it implements __len__ and such)
            with self.subTest(num_args=num_args, input='list'):
                chunks = list(chunk_tasks(range(num_args), chunk_size=chunk_size))
                for chunk in chunks[:-1]:
                    self.assertEqual(len(chunk), chunk_size)
                self.assertLessEqual(len(chunks[-1]), chunk_size)
                self.assertEqual(list(range(num_args)), list(chain.from_iterable(chunks)))

            # Test for an actual generator (range does not really behave like one)
            with self.subTest(num_args=num_args, input='generator/iterator'):
                chunks = list(chunk_tasks(iter(range(num_args)), chunk_size=chunk_size))
                for chunk in chunks[:-1]:
                    self.assertEqual(len(chunk), chunk_size)
                self.assertLessEqual(len(chunks[-1]), chunk_size)
                self.assertEqual(list(range(num_args)), list(chain.from_iterable(chunks)))


class ApplyNumpyChunkingTest(unittest.TestCase):

    """
    This function simply calls other, already tested, functions in succession. We do test the individual parameter
    influence, but interactions between them are skipped
    """

    def setUp(self):
        self.test_data_numpy = np.random.rand(100, 2)

    def test_iterable_len(self):
        """
        Test that iterable_len is adhered to. When iterable_len < len(input) it should reduce the input size. If higher
        or None it should take the entire input
        """
        for iterable_len, expected_size in [(5, 5), (150, 100), (None, 100)]:
            with self.subTest(iterable_len=iterable_len):
                iterable_of_args, iterable_len_, chunk_size, n_splits = apply_numpy_chunking(
                    self.test_data_numpy, iterable_len=iterable_len, n_splits=1
                )

                # Materialize generator and test contents
                iterable_of_args = list(iterable_of_args)
                self.assertEqual(len(iterable_of_args), 1)
                self.assertIsInstance(iterable_of_args[0][0], np.ndarray)
                np.testing.assert_array_equal(iterable_of_args[0][0], self.test_data_numpy[:expected_size])

                # Test other output
                self.assertEqual(iterable_len_, 1)
                self.assertEqual(chunk_size, 1)
                self.assertIsNone(n_splits)

    def test_chunk_size(self):
        """
        Test that chunk_size works as expected. Note that chunk_size trumps n_splits
        """
        for chunk_size, expected_n_chunks in [(1, 100), (3, 34), (200, 1), (None, 1)]:
            with self.subTest(chunk_size=chunk_size):
                iterable_of_args, iterable_len, chunk_size_, n_splits = apply_numpy_chunking(
                    self.test_data_numpy, chunk_size=chunk_size, n_splits=1
                )

                # Materialize generator and test contents. The chunks should be of size chunk_size (expect for the last
                # chunk which can be smaller)
                iterable_of_args = list(iterable_of_args)
                self.assertEqual(len(iterable_of_args), expected_n_chunks)
                chunk_size = chunk_size or 100
                for chunk_idx, chunk in enumerate(iterable_of_args):
                    self.assertIsInstance(chunk[0], np.ndarray)
                    np.testing.assert_array_equal(chunk[0], self.test_data_numpy[chunk_idx * chunk_size:
                                                                                 (chunk_idx + 1) * chunk_size])

                # Test other output
                self.assertEqual(iterable_len, expected_n_chunks)
                self.assertEqual(chunk_size_, 1)
                self.assertIsNone(n_splits)

    def test_n_splits(self):
        """
        Test that n_splits works as expected.
        """
        for n_splits, expected_n_chunks in [(1, 1), (3, 3), (150, 100)]:
            with self.subTest(n_splits=n_splits):
                iterable_of_args, iterable_len, chunk_size, n_splits_ = apply_numpy_chunking(
                    self.test_data_numpy, n_splits=n_splits
                )

                # Materialize generator and test contents. We simply test if every row of the original input occurs in
                # the chunks
                iterable_of_args = list(iterable_of_args)
                self.assertEqual(len(iterable_of_args), expected_n_chunks)
                offset = 0
                for chunk in iterable_of_args:
                    self.assertIsInstance(chunk[0], np.ndarray)
                    np.testing.assert_array_equal(chunk[0], self.test_data_numpy[offset:offset + len(chunk[0])])
                    offset += len(chunk[0])
                self.assertEqual(offset, 100)

                # Test other output
                self.assertEqual(iterable_len, expected_n_chunks)
                self.assertEqual(chunk_size, 1)
                self.assertIsNone(n_splits_)

        # chunk_size and n_splits can't be both None
        with self.subTest(n_splits=None), self.assertRaises(ValueError):
            iterable_of_args, *_ = apply_numpy_chunking(self.test_data_numpy, n_splits=None)
            list(iterable_of_args)

    def test_n_jobs(self):
        """
        Test that n_jobs works as expected. When chunk_size and n_splits are both None, n_jobs * 4 is passed on as
        n_splits
        """
        for n_jobs, expected_n_chunks in [(1, 4), (3, 12), (40, 100), (150, 100)]:
            with self.subTest(n_jobs=n_jobs):
                iterable_of_args, iterable_len, chunk_size, n_splits_ = apply_numpy_chunking(
                    self.test_data_numpy, n_jobs=n_jobs
                )

                # Materialize generator and test contents. We simply test if every row of the original input occurs in
                # the chunks
                iterable_of_args = list(iterable_of_args)
                self.assertEqual(len(iterable_of_args), expected_n_chunks)
                offset = 0
                for chunk in iterable_of_args:
                    self.assertIsInstance(chunk[0], np.ndarray)
                    np.testing.assert_array_equal(chunk[0], self.test_data_numpy[offset:offset + len(chunk[0])])
                    offset += len(chunk[0])
                self.assertEqual(offset, 100)

                # Test other output
                self.assertEqual(iterable_len, expected_n_chunks)
                self.assertEqual(chunk_size, 1)
                self.assertIsNone(n_splits_)


class GetNChunksTest(unittest.TestCase):

    def setUp(self):
        self.test_data = [1, 2, 3, 5, 6, 9, 37, 42, 1337, 0, 3, 5, 0]
        self.test_data_numpy = np.random.rand(100, 2)

    def test_everything_none(self):
        """
        When everything is None we should use cpu_count * 4 as number of splits. We have to take the number of tasks
        into account
        """
        with self.subTest(input='list'):
            self.assertEqual(get_n_chunks(self.test_data, iterable_len=None, chunk_size=None, n_splits=None,
                                          n_jobs=None), min(13, cpu_count() * 4))
        with self.subTest(input='numpy'):
            self.assertEqual(get_n_chunks(self.test_data_numpy, iterable_len=None, chunk_size=None, n_splits=None,
                                          n_jobs=None), min(100, cpu_count() * 4))

    def test_smaller_iterable_len(self):
        """
        Test iterable_len, where iterable_len < len(input)
        """
        with self.subTest(input='list'):
            self.assertEqual(get_n_chunks(self.test_data, iterable_len=5, chunk_size=None, n_splits=None, n_jobs=None),
                             min(5, cpu_count() * 4))
        with self.subTest(input='numpy'):
            self.assertEqual(get_n_chunks(self.test_data_numpy, iterable_len=5, chunk_size=None, n_splits=None,
                                          n_jobs=None), min(5, cpu_count() * 4))
        with self.subTest(input='generator/iterator'):
            self.assertEqual(get_n_chunks(iter(self.test_data), iterable_len=5, chunk_size=None, n_splits=None,
                                          n_jobs=None), min(5, cpu_count() * 4))

    def test_larger_iterable_len(self):
        """
        Test iterable_len, where iterable_len > len(input). Should ignores iterable_len when actual number of tasks is
        less, except when we use the data_generator function, in which case we cannot determine the actual number of
        elements.
        """
        with self.subTest(input='list'):
            self.assertEqual(get_n_chunks(self.test_data, iterable_len=25, chunk_size=None, n_splits=None, n_jobs=None),
                             min(13, cpu_count() * 4))
        with self.subTest(input='numpy'):
            self.assertEqual(get_n_chunks(self.test_data_numpy, iterable_len=125, chunk_size=None, n_splits=None,
                                          n_jobs=None), min(100, cpu_count() * 4))
        with self.subTest(input='generator/iterator'):
            self.assertEqual(get_n_chunks(iter(self.test_data), iterable_len=25, chunk_size=None, n_splits=None,
                                          n_jobs=None), min(25, cpu_count() * 4))

    def test_chunk_size(self):
        """
        Test chunk_size
        """
        for chunk_size, expected_n_chunks in [(1, 13), (3, 5)]:
            with self.subTest(input='list', chunk_size=chunk_size, expected_n_chunks=expected_n_chunks):
                self.assertEqual(get_n_chunks(self.test_data, iterable_len=None, chunk_size=chunk_size, n_splits=None,
                                              n_jobs=None), expected_n_chunks)

        for chunk_size, expected_n_chunks in [(1, 100), (3, 34)]:
            with self.subTest(input='list', chunk_size=chunk_size, expected_n_chunks=expected_n_chunks):
                self.assertEqual(get_n_chunks(self.test_data_numpy, iterable_len=None, chunk_size=chunk_size,
                                              n_splits=None, n_jobs=None), expected_n_chunks)

    def test_n_splits(self):
        """
        Test n_splits. n_jobs shouldn't have any influence
        """
        for n_splits, n_jobs in product([1, 6], [None, 2, 8]):
            with self.subTest(input='list', n_splits=n_splits, n_jobs=n_jobs):
                self.assertEqual(get_n_chunks(self.test_data, iterable_len=None, chunk_size=None, n_splits=n_splits,
                                              n_jobs=n_jobs), n_splits)

            with self.subTest(input='numpy', n_splits=n_splits, n_jobs=n_jobs):
                self.assertEqual(get_n_chunks(self.test_data_numpy, iterable_len=None, chunk_size=None,
                                              n_splits=n_splits, n_jobs=n_jobs), n_splits)

    def test_n_jobs(self):
        """
        When everything is None except n_jobs we should use n_jobs * 4 as number of splits. Again, taking into account
        the number of tasks
        """
        for n_jobs in [1, 6]:
            with self.subTest(input='list', n_jobs=n_jobs):
                self.assertEqual(get_n_chunks(self.test_data, iterable_len=None, chunk_size=None, n_splits=None,
                                              n_jobs=n_jobs), min(4 * n_jobs, len(self.test_data)))

            with self.subTest(input='numpy', n_jobs=n_jobs):
                self.assertEqual(get_n_chunks(self.test_data_numpy, iterable_len=None, chunk_size=None, n_splits=None,
                                              n_jobs=n_jobs), min(4 * n_jobs, len(self.test_data_numpy)))

    def test_chunk_size_priority_over_n_splits(self):
        """
        chunk_size should have priority over n_splits
        """
        with self.subTest(input='list', chunk_size=1, n_splits=6):
            self.assertEqual(get_n_chunks(self.test_data, iterable_len=None, chunk_size=1, n_splits=6, n_jobs=None), 13)
        with self.subTest(input='numpy', chunk_size=1, n_splits=6):
            self.assertEqual(get_n_chunks(self.test_data_numpy, iterable_len=None, chunk_size=1, n_splits=6,
                                          n_jobs=None), 100)

        with self.subTest(input='list', chunk_size=3, n_splits=3):
            self.assertEqual(get_n_chunks(self.test_data, iterable_len=None, chunk_size=3, n_splits=3, n_jobs=None), 5)
        with self.subTest(input='numpy', chunk_size=3, n_splits=3):
            self.assertEqual(get_n_chunks(self.test_data_numpy, iterable_len=None, chunk_size=3, n_splits=3,
                                          n_jobs=None), 34)

    def test_generator_input_with_no_iterable_len_raises(self):
        """
        When working with generators the iterable_len should be provided (the working examples are already tested above)
        """
        for chunk_size, n_splits, n_jobs in product([None, 1, 3], [None, 1, 3], [None, 1, 3]):
            with self.subTest(chunk_size=chunk_size, n_splits=n_splits, n_jobs=n_jobs), self.assertRaises(ValueError):
                get_n_chunks(iter(self.test_data), iterable_len=None, chunk_size=chunk_size, n_splits=n_splits,
                             n_jobs=n_jobs)


class MakeSingleArgumentsTest(unittest.TestCase):

    def test_make_single_arguments(self):
        """
        Tests the make_single_arguments function for different inputs
        """
        # Test for some different inputs
        for (args_in, args_out), generator in product(
                [(['a', 'c', 'b', 'd'], [('a',), ('c',), ('b',), ('d',)]),
                 ([1, 2, 3, 4, 5], [(1,), (2,), (3,), (4,), (5,)]),
                 ([(True,), (False,), (None,)], [((True,),), ((False,),), ((None,),)])],
                [False, True]
        ):
            # Transform
            args_transformed = make_single_arguments((arg for arg in args_in) if generator else args_in,
                                                     generator=generator)

            # Check type
            self.assertTrue(isinstance(args_transformed, types.GeneratorType if generator else list))

            # Check contents
            self.assertEqual(list(args_transformed), args_out)


class FormatSecondsTest(unittest.TestCase):

    def test_none_input(self):
        """
        When the input is None it should return an empty string
        """
        for with_milliseconds in [False, True]:
            with self.subTest(with_milliseconds=with_milliseconds):
                self.assertEqual(format_seconds(None, with_milliseconds=with_milliseconds), '')

    def test_without_milliseconds(self):
        """
        Test output without milliseconds
        """
        for seconds, expected_output in [(0, '0:00:00'), (1, '0:00:01'), (1.337, '0:00:01'), (2.9, '0:00:02'),
                                         (123456.78901234, '1 day, 10:17:36')]:
            with self.subTest(seconds=seconds):
                self.assertEqual(format_seconds(seconds, with_milliseconds=False), expected_output)

    def test_with_milliseconds(self):
        """
        Test output with milliseconds. Only shows them when they're actually needed.
        """
        for seconds, expected_output in [(0, '0:00:00'), (1, '0:00:01'), (1.337, '0:00:01.337'), (2.9, '0:00:02.900'),
                                         (123456.78901234, '1 day, 10:17:36.789')]:
            with self.subTest(seconds=seconds):
                self.assertEqual(format_seconds(seconds, with_milliseconds=True), expected_output)


class TimeItTest(unittest.TestCase):

    def test_array_storage(self):
        """
        TimeIt should write to the correct idx in the cum_time_array container. The max_time_array is a min-heap
        container, so the lowest value is stored at index 0. The single highest value in this case is stored at index 2
        """
        for array_idx in range(5):
            cum_time_array = [0, 0, 0, 0, 0]
            max_time_array = [(0, ''), (0, ''), (0, ''), (0, ''), (0, '')]
            MockDatetimeNow.RETURN_VALUES = [datetime(1970, 1, 1, 0, 0, 0, 0),
                                             datetime(1970, 1, 1, 0, 0, 4, 200000)]
            MockDatetimeNow.CURRENT_IDX = 0
            with self.subTest(array_idx=array_idx), patch('mpire.utils.datetime', new=MockDatetimeNow), \
                    TimeIt(cum_time_array, array_idx, max_time_array):
                pass
            self.assertListEqual([t for idx, t in enumerate(cum_time_array) if idx != array_idx], [0, 0, 0, 0])
            self.assertListEqual([t for idx, t in enumerate(max_time_array) if idx != 2],
                                 [(0, ''), (0, ''), (0, ''), (0, '')])
            self.assertEqual(cum_time_array[array_idx], 4.2)
            self.assertGreaterEqual(max_time_array[2], (4.2, None))

    def test_cum_time(self):
        """
        Using TimeIt multiple times should increase the cum_time_array
        """
        # These return values are used by TimeIt in order: start, end, start, end, ... So the first time the duration
        # will be 1 second, then 2 seconds, and 3 seconds.
        MockDatetimeNow.RETURN_VALUES = [datetime(1970, 1, 1, 0, 0, 0, 0),
                                         datetime(1970, 1, 1, 0, 0, 1, 0),
                                         datetime(1970, 1, 1, 0, 0, 0, 0),
                                         datetime(1970, 1, 1, 0, 0, 2, 0),
                                         datetime(1970, 1, 1, 0, 0, 0, 0),
                                         datetime(1970, 1, 1, 0, 0, 3, 0)]
        MockDatetimeNow.CURRENT_IDX = 0
        cum_time_array = [0]
        with patch('mpire.utils.datetime', new=MockDatetimeNow):
            with TimeIt(cum_time_array, 0):
                pass
            self.assertEqual(cum_time_array[0], 1.0)

            with TimeIt(cum_time_array, 0):
                pass
            self.assertEqual(cum_time_array[0], 3.0)

            with TimeIt(cum_time_array, 0):
                pass
            self.assertEqual(cum_time_array[0], 6.0)

    def test_max_time(self):
        """
        Using TimeIt multiple times should store the max duration value in the max_time_array using heapq. There's only
        room for the highest 5 values, while it is called 6 times. The smallest duration shouldn't be present.
        """
        # These return values are used by TimeIt in order: start, end, start, end, ... So the first time the duration
        # will be 1 second, then 2 seconds, 3 seconds, 3 seconds again, 0.5 seconds, and 10 seconds.
        MockDatetimeNow.RETURN_VALUES = [datetime(1970, 1, 1, 0, 0, 0, 0),
                                         datetime(1970, 1, 1, 0, 0, 1, 0),
                                         datetime(1970, 1, 1, 0, 0, 0, 0),
                                         datetime(1970, 1, 1, 0, 0, 2, 0),
                                         datetime(1970, 1, 1, 0, 0, 0, 0),
                                         datetime(1970, 1, 1, 0, 0, 3, 0),
                                         datetime(1970, 1, 1, 0, 0, 0, 0),
                                         datetime(1970, 1, 1, 0, 0, 3, 0),
                                         datetime(1970, 1, 1, 0, 0, 0, 0),
                                         datetime(1970, 1, 1, 0, 0, 0, 500000),
                                         datetime(1970, 1, 1, 0, 0, 0, 0),
                                         datetime(1970, 1, 1, 0, 0, 10, 0)]
        MockDatetimeNow.CURRENT_IDX = 0
        cum_time_array = [0]
        max_time_array = [(0, ''), (0, ''), (0, ''), (0, ''), (0, '')]
        with patch('mpire.utils.datetime', new=MockDatetimeNow):
            for _ in range(6):
                with TimeIt(cum_time_array, 0, max_time_array):
                    pass
            self.assertListEqual(max_time_array, [(1.0, None), (2.0, None), (10.0, None), (3.0, None), (3.0, None)])

    def test_format_args(self):
        """
        The format args func should be called when provided
        """
        for format_func, formatted in [(lambda: "1", "1"), (lambda: 2, 2), (lambda: "foo", "foo")]:
            # These return values are used by TimeIt in order: start, end, start, end, ... So the first time the
            # duration will be 1 second, then 2 seconds, and 3 seconds.
            MockDatetimeNow.RETURN_VALUES = [datetime(1970, 1, 1, 0, 0, 0, 0),
                                             datetime(1970, 1, 1, 0, 0, 1, 0),
                                             datetime(1970, 1, 1, 0, 0, 0, 0),
                                             datetime(1970, 1, 1, 0, 0, 2, 0),
                                             datetime(1970, 1, 1, 0, 0, 0, 0),
                                             datetime(1970, 1, 1, 0, 0, 3, 0)]
            MockDatetimeNow.CURRENT_IDX = 0
            with self.subTest(format_func=format_func), patch('mpire.utils.datetime', new=MockDatetimeNow):
                cum_time_array = [0]
                max_time_array = [(0, ''), (0, '')]
                for _ in range(3):
                    with TimeIt(cum_time_array, 0, max_time_array, format_func):
                        pass

                # The heapq only had room for two entries. The highest durations should be kept
                self.assertListEqual(max_time_array, [(2.0, formatted), (3.0, formatted)])



import _thread
import ctypes
import multiprocess as mp_dill
import multiprocessing as mp
import multiprocessing.synchronize
import queue
import threading
import unittest
from datetime import datetime
from itertools import product
from unittest.mock import patch

from mpire.comms import NON_LETHAL_POISON_PILL, POISON_PILL, WorkerComms
from mpire.context import DEFAULT_START_METHOD, FORK_AVAILABLE, MP_CONTEXTS
from tests.utils import MockDatetimeNow


class WorkerCommsTest(unittest.TestCase):

    def test_init_comms(self):
        """
        Test if initializing/resetting the comms is done properly
        """
        test_ctx = [(MP_CONTEXTS['mp_dill']['spawn'], True, False, MP_CONTEXTS['mp_dill'][DEFAULT_START_METHOD]),
                    (MP_CONTEXTS['threading'], False, True, MP_CONTEXTS['mp'][DEFAULT_START_METHOD])]
        if FORK_AVAILABLE:
            test_ctx.extend([(MP_CONTEXTS['mp_dill']['fork'], True, False, MP_CONTEXTS['mp_dill']['fork']),
                             (MP_CONTEXTS['mp']['forkserver'], False, False, MP_CONTEXTS['mp']['fork'])])

        for n_jobs, (ctx, use_dill, using_threading, expected_ctx_for_threading) in product([1, 2, 4], test_ctx):
            comms = WorkerComms(ctx, n_jobs, use_dill, using_threading)
            self.assertEqual(comms.ctx, ctx)
            self.assertEqual(comms.ctx_for_threading, expected_ctx_for_threading)
            self.assertEqual(comms.n_jobs, n_jobs)
            self.assertEqual(comms.using_threading, using_threading)

            expected_mp = mp_dill if use_dill else mp

            with self.subTest('__init__ called', n_jobs=n_jobs, using_threading=using_threading):
                self.assertIsInstance(comms._keep_order, type(ctx.Event()))
                self.assertFalse(comms._keep_order.is_set())
                self.assertIsNone(comms._task_queues)
                self.assertIsNone(comms._task_idx)
                self.assertIsNone(comms._last_completed_task_worker_id)
                self.assertIsNone(comms._results_queue)
                self.assertListEqual(comms._exit_results_queues, [])
                self.assertIsNone(comms._all_exit_results_obtained)
                self.assertIsNone(comms._worker_done_array)
                self.assertIsNone(comms._workers_dead)
                self.assertIsNone(comms._exception_queue)
                self.assertIsInstance(comms.exception_lock,
                                      _thread.LockType if using_threading else expected_mp.synchronize.Lock)
                self.assertIsInstance(comms._exception_thrown, expected_mp.synchronize.Event)
                self.assertIsInstance(comms._kill_signal_received, expected_mp.synchronize.Event)
                self.assertFalse(comms._exception_thrown.is_set())
                self.assertFalse(comms._kill_signal_received.is_set())
                self.assertIsNone(comms._task_completed_queue)
                self.assertIsNone(comms._all_exit_results_obtained)

            with self.subTest('without initial values', n_jobs=n_jobs, using_threading=using_threading,
                              has_worker_exit=False, has_progress_bar=False):
                comms.init_comms(has_worker_exit=False, has_progress_bar=False)
                self.assertFalse(comms._keep_order.is_set())
                self.assertEqual(len(comms._task_queues), n_jobs)
                for q in comms._task_queues:
                    self.assertIsInstance(q, expected_mp.queues.JoinableQueue)
                self.assertIsNone(comms._last_completed_task_worker_id)
                self.assertIsInstance(comms._results_queue, expected_mp.queues.JoinableQueue)
                self.assertListEqual(comms._exit_results_queues, [])
                self.assertIsInstance(comms._worker_done_array, ctypes.Array)
                self.assertEqual(len(comms._workers_dead), n_jobs)
                for worker_dead in comms._workers_dead:
                    self.assertIsInstance(worker_dead,
                                          threading.Event if using_threading else expected_mp.synchronize.Event)
                    self.assertTrue(worker_dead.is_set())
                self.assertIsInstance(comms._exception_queue, expected_mp.queues.JoinableQueue)
                self.assertFalse(comms._exception_thrown.is_set())
                self.assertFalse(comms._kill_signal_received.is_set())
                self.assertIsNone(comms._task_completed_queue)
                self.assertIsNone(comms._progress_bar_complete)

                # Basic sanity checks for the values
                self.assertEqual(comms._task_idx, 0)
                self.assertEqual(list(comms._worker_done_array), [False for _ in range(n_jobs)])

            with self.subTest('without initial values', n_jobs=n_jobs, using_threading=using_threading,
                              has_worker_exit=True, has_progress_bar=True):
                comms.init_comms(has_worker_exit=True, has_progress_bar=True)
                self.assertEqual(len(comms._exit_results_queues), n_jobs)
                for q in comms._exit_results_queues:
                    self.assertIsInstance(q, expected_mp.queues.JoinableQueue)
                self.assertIsInstance(comms._all_exit_results_obtained,
                                      threading.Event if using_threading else expected_mp.synchronize.Event)
                self.assertFalse(comms._all_exit_results_obtained.is_set())
                self.assertIsInstance(comms._task_completed_queue, expected_mp.queues.JoinableQueue)
                self.assertIsInstance(comms._progress_bar_complete, expected_mp.synchronize.Event)
                self.assertFalse(comms._progress_bar_complete.is_set())

            # Set some values so we can test if the containers will be properly resetted
            comms._keep_order.set()
            comms._task_idx = 4
            comms._last_completed_task_worker_id = 2
            comms._worker_done_array[:] = [False, True, False, True][:n_jobs]
            [worker_dead.clear() for worker_dead in comms._workers_dead]
            comms._exception_thrown.set()
            comms._kill_signal_received.set()

            # Note that threading doesn't have a JoinableQueue, so they're taken from multiprocessing
            with self.subTest('with initial values', n_jobs=n_jobs, using_threading=using_threading,
                              has_worker_exit=False, has_progress_bar=False):
                comms.init_comms(has_worker_exit=False, has_progress_bar=False)
                self.assertEqual(len(comms._task_queues), n_jobs)
                for q in comms._task_queues:
                    self.assertIsInstance(q, expected_mp.queues.JoinableQueue)
                self.assertIsNone(comms._last_completed_task_worker_id)
                self.assertIsInstance(comms._results_queue, expected_mp.queues.JoinableQueue)
                self.assertListEqual(comms._exit_results_queues, [])
                self.assertIsInstance(comms._worker_done_array, ctypes.Array)
                self.assertEqual(len(comms._workers_dead), n_jobs)
                for worker_dead in comms._workers_dead:
                    self.assertIsInstance(worker_dead,
                                          threading.Event if using_threading else expected_mp.synchronize.Event)
                    self.assertTrue(worker_dead.is_set())
                self.assertIsInstance(comms._exception_queue, expected_mp.queues.JoinableQueue)
                self.assertFalse(comms._exception_thrown.is_set())
                self.assertFalse(comms._kill_signal_received.is_set())
                self.assertIsNone(comms._task_completed_queue)

                # Some variables are not reset by this function, but are reset otherwise
                self.assertIsInstance(comms._keep_order,
                                      threading.Event if using_threading else expected_mp.synchronize.Event)
                self.assertTrue(comms._keep_order.is_set())

                # Basic sanity checks for the values
                self.assertEqual(comms._task_idx, 0)
                self.assertEqual(list(comms._worker_done_array), [False for _ in range(n_jobs)])

    def test_progress_bar(self):
        """
        Test progress bar related functions
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 2, False, False)

        # Has progress bar
        self.assertFalse(comms.has_progress_bar())
        comms.init_comms(False, True)
        self.assertTrue(comms.has_progress_bar())

        # Nothing available yet
        with self.assertRaises(queue.Empty):
            comms._task_completed_queue.get(block=False)

        MockDatetimeNow.RETURN_VALUES = [datetime(1970, 1, 1, 0, 0, 0, 0),
                                         datetime(1970, 1, 1, 0, 0, 0, 0),
                                         datetime(1970, 1, 1, 0, 0, 0, 0),
                                         datetime(1970, 1, 1, 0, 0, 0, 0)]
        MockDatetimeNow.CURRENT_IDX = 0

        # 3 task done, but not enough time has passed to send the update
        last_updated = datetime(1970, 1, 1, 0, 0, 0, 0)
        n_tasks_completed = 0
        with patch('mpire.comms.datetime', new=MockDatetimeNow):
            for n in range(1, 4):
                last_updated, n_tasks_completed = comms.task_completed_progress_bar(last_updated, n_tasks_completed,
                                                                                    force_update=False)
                self.assertEqual(n_tasks_completed, n)
        with self.assertRaises(queue.Empty):
            comms._task_completed_queue.get(block=False)

        # Not enough time has passed, but we'll force the update. Number of tasks done should still be 3
        with patch('mpire.comms.datetime', new=MockDatetimeNow):
            last_updated, n_tasks_completed = comms.task_completed_progress_bar(last_updated, n_tasks_completed,
                                                                                force_update=True)
        self.assertEqual(comms.get_tasks_completed_progress_bar(), (3, True))
        self.assertEqual(last_updated, datetime(1970, 1, 1, 0, 0, 0, 0))
        self.assertEqual(n_tasks_completed, 0)
        comms.task_done_progress_bar()

        # 4 tasks already done and another 4 tasks done. Enough time should've passed for each update call, except the
        # second
        MockDatetimeNow.RETURN_VALUES = [datetime(1970, 1, 1, 0, 1, 0, 0),
                                         datetime(1970, 1, 1, 0, 1, 0, 0),
                                         datetime(1970, 1, 1, 0, 3, 0, 0),
                                         datetime(1970, 1, 1, 0, 4, 0, 0)]
        MockDatetimeNow.CURRENT_IDX = 0
        last_updated = datetime(1970, 1, 1, 0, 0, 0, 0)
        n_tasks_completed = 4
        with patch('mpire.comms.datetime', new=MockDatetimeNow):
            for _ in range(4):
                last_updated, n_tasks_completed = comms.task_completed_progress_bar(last_updated, n_tasks_completed,
                                                                                    force_update=False)
                self.assertEqual(last_updated, MockDatetimeNow.RETURN_VALUES[MockDatetimeNow.CURRENT_IDX - 1])
        self.assertListEqual([comms.get_tasks_completed_progress_bar() for _ in range(3)],
                             [(5, True), (2, True), (1, True)])
        self.assertEqual(last_updated, datetime(1970, 1, 1, 0, 4, 0, 0))
        self.assertEqual(n_tasks_completed, 0)
        [comms.task_done_progress_bar() for _ in range(3)]

        # Add poison pill
        comms.add_progress_bar_poison_pill()
        self.assertEqual(comms.get_tasks_completed_progress_bar(), (POISON_PILL, True))
        comms.task_done_progress_bar()

        # Set exception
        comms.set_exception_thrown()
        self.assertEqual(comms.get_tasks_completed_progress_bar(), (POISON_PILL, False))
        comms._exception_thrown.clear()

        # Set kill signal received
        comms.set_kill_signal_received()
        self.assertEqual(comms.get_tasks_completed_progress_bar(), (POISON_PILL, False))
        comms._kill_signal_received.clear()

        # Should be joinable. When using keep_alive, it should still be open
        comms.join_progress_bar_task_completed_queue(keep_alive=True)
        comms.task_completed_progress_bar(force_update=True)
        comms.get_tasks_completed_progress_bar()
        comms.task_done_progress_bar()
        comms.join_progress_bar_task_completed_queue(keep_alive=False)

        # Depending on Python version it either throws AssertionError or ValueError
        with self.assertRaises((AssertionError, ValueError)):
            comms.task_completed_progress_bar(force_update=True)

    def test_progress_bar_complete(self):
        """
        Test progress bar complete related functions
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 2, False, False)
        comms.init_comms(False, True)

        self.assertFalse(comms._progress_bar_complete.is_set())
        comms.set_progress_bar_complete()
        self.assertTrue(comms._progress_bar_complete.is_set())

        with patch.object(comms._progress_bar_complete, 'wait') as p:
            comms.wait_until_progress_bar_is_complete()
            self.assertEqual(p.call_count, 1)

    def test_keep_order(self):
        """
        Test keep_order related functions
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 2, False, False)

        self.assertFalse(comms.keep_order())
        comms.set_keep_order()
        self.assertTrue(comms.keep_order())
        comms.clear_keep_order()
        self.assertFalse(comms.keep_order())

    def test_tasks(self):
        """
        Test task related functions
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 3, False, False)
        comms.init_comms(False, False)

        # Nothing available yet
        for worker_id in range(3):
            with self.assertRaises(queue.Empty):
                comms._task_queues[worker_id].get(block=False)

        # Add a few tasks. As no worker has completed tasks yet, it should give the task to workers in order
        comms.add_task(12)
        comms.add_task('hello world')
        comms.add_task({'foo': 'bar'})
        comms.add_task({'foo': 'baz'})
        comms.add_task(34.43)
        comms.add_task(datetime(2000, 1, 1, 1, 2, 3))
        tasks = []
        for worker_id in [0, 1, 2, 0, 1, 2]:
            tasks.append(comms.get_task(worker_id))
            comms.task_done(worker_id)
        self.assertListEqual(tasks, [12, 'hello world', {'foo': 'bar'}, {'foo': 'baz'}, 34.43,
                                     datetime(2000, 1, 1, 1, 2, 3)])

        # When workers have completed tasks, it depends on the last one who gets the new task. After giving a task to
        # that worker the worker ID that last completed a task is reset again. So, it should continue with the normal
        # order
        comms.init_comms(False, False)
        comms.add_task(12)
        comms.add_task('hello world')
        comms.add_task({'foo': 'bar'})
        comms._last_completed_task_worker_id = 2
        comms.add_task({'foo': 'baz'})
        comms._last_completed_task_worker_id = 1
        comms.add_task(34.43)
        comms._last_completed_task_worker_id = 0
        comms.add_task(datetime(2000, 1, 1, 1, 2, 3))
        comms._last_completed_task_worker_id = 2
        comms.add_task('123')
        comms.add_task(123)
        comms.add_task(1337)
        tasks = []
        for worker_id in [0, 1, 2, 2, 1, 0, 2, 0, 1]:
            tasks.append(comms.get_task(worker_id))
            comms.task_done(worker_id)
        self.assertListEqual(tasks, [12, 'hello world', {'foo': 'bar'}, {'foo': 'baz'}, 34.43,
                                     datetime(2000, 1, 1, 1, 2, 3), '123', 123, 1337])

        # Throw in an exception. Should return None
        comms.set_exception_thrown()
        for worker_id in range(3):
            self.assertIsNone(comms.get_task(worker_id))

        # Should be joinable
        comms.join_task_queues()

    def test_results(self):
        """
        Test results related functions
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 2, False, False)
        comms.init_comms(False, False)

        # Nothing available yet
        with self.assertRaises(queue.Empty):
            comms._results_queue.get(block=False)

        # Add a few results. Note that `get_results` calls `task_done`
        comms.add_results(0, 12)
        comms.add_results(1, 'hello world')
        comms.add_results(1, {'foo': 'bar'})
        comms.add_results(0, '123')
        self.assertEqual(comms.get_results(), 12)
        self.assertEqual(comms._last_completed_task_worker_id, 0)
        self.assertEqual(comms.get_results(), 'hello world')
        self.assertEqual(comms._last_completed_task_worker_id, 1)
        self.assertEqual(comms.get_results(), {'foo': 'bar'})
        self.assertEqual(comms._last_completed_task_worker_id, 1)
        self.assertEqual(comms.get_results(), '123')
        self.assertEqual(comms._last_completed_task_worker_id, 0)

        # Should be joinable. When using keep_alive, it should still be open
        comms.join_results_queues(keep_alive=True)
        comms.add_results(0, 12)
        comms.get_results()
        comms.join_results_queues(keep_alive=False)

        # Depending on Python version it either throws AssertionError or ValueError
        with self.assertRaises((AssertionError, ValueError)):
            comms.add_results(0, 12)

    def test_exit_results(self):
        """
        Test exit results related functions
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 3, False, False)
        comms.init_comms(True, False)

        # Nothing available yet
        for worker_id in range(3):
            with self.assertRaises(queue.Empty):
                comms.get_exit_results(worker_id, block=False)

        # Add a few results. Note that `get_exit_results` calls `task_done`
        for worker_id in range(3):
            comms.add_exit_results(worker_id, worker_id)
            comms.add_exit_results(worker_id, 'hello world')
            comms.add_exit_results(worker_id, {'foo': 'bar'})
        self.assertListEqual([comms.get_exit_results(worker_id=0) for _ in range(3)],
                             [0, 'hello world', {'foo': 'bar'}])
        self.assertListEqual([comms.get_exit_results(worker_id=1) for _ in range(3)],
                             [1, 'hello world', {'foo': 'bar'}])
        self.assertListEqual([comms.get_exit_results(worker_id=2) for _ in range(3)],
                             [2, 'hello world', {'foo': 'bar'}])

        # When an exception is thrown it should return directly. However, when block=False it should still get it. We
        # use a while loop because we need to wait until the queue is ready. This block=False trick is used in
        # terminate().
        comms.add_exit_results(0, 'hello world')
        comms.set_exception_thrown()
        self.assertIsNone(comms.get_exit_results(0))
        while True:
            try:
                self.assertEqual(comms.get_exit_results(0, block=False), 'hello world')
                break
            except queue.Empty:
                pass
        comms._exception_thrown.clear()

        # Should be joinable. When using keep_alive, it should still be open
        comms.join_results_queues(keep_alive=True)
        for worker_id in range(3):
            comms.add_exit_results(worker_id, 'hello world')
            comms.get_exit_results(worker_id)
        comms.join_results_queues(keep_alive=False)

        # Depending on Python version it either throws AssertionError or ValueError
        for worker_id in range(3):
            with self.assertRaises((AssertionError, ValueError)):
                comms.add_exit_results(worker_id, 'hello world')

    def test_exit_results_all_workers(self):
        """
        Test exit results related functions
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 4, False, False)
        comms.init_comms(True, False)

        # Add a few results. Every worker will always have a return value (even if it's the implicit None). Note that
        # `get_exit_results` calls `task_done`
        for worker_id in range(3):
            comms.add_exit_results(worker_id, worker_id)
        comms.add_exit_results(3, None)
        self.assertListEqual(comms.get_exit_results_all_workers(), [0, 1, 2, None])

        # Should be joinable
        comms.join_results_queues()

    def test_exit_results_all_workers_exception_thrown(self):
        """
        Test exit results related functions. When an exception occurred, it should return an empty list
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 3, False, False)
        comms.init_comms(True, False)

        # Add a few results.
        for worker_id in range(3):
            comms.add_exit_results(worker_id, worker_id)

        # Set exception
        comms.set_exception_thrown()

        # Should return empty list
        self.assertListEqual(comms.get_exit_results_all_workers(), [])

        # Drain and join
        comms._exception_thrown.clear()
        comms.get_exit_results_all_workers()
        comms.join_results_queues()

    def test_all_exit_results_obtained(self):
        """
        Test all exit results obtained related functions
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 2, False, False)
        comms.init_comms(True, False)

        self.assertFalse(comms._all_exit_results_obtained.is_set())
        comms.set_all_exit_results_obtained()
        self.assertTrue(comms._all_exit_results_obtained.is_set())

        with patch.object(comms._all_exit_results_obtained, 'wait') as p:
            comms.wait_until_all_exit_results_obtained()
            self.assertEqual(p.call_count, 1)

    def test_exceptions(self):
        """
        Test exception related functions
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 2, False, False)
        comms.init_comms(False, False)

        # Nothing available yet
        with self.assertRaises(queue.Empty):
            comms._exception_queue.get(block=False)

        # Add a few exceptions
        comms.add_exception(TypeError, 'TypeError')
        comms.add_exception(ValueError, 'ValueError')
        comms.add_exception(RuntimeError, 'RuntimeError')
        self.assertListEqual([comms.get_exception() for _ in range(3)],
                             [(TypeError, 'TypeError'), (ValueError, 'ValueError'), (RuntimeError, 'RuntimeError')])
        [comms.task_done_exception() for _ in range(3)]

        # Add poison pill
        comms.add_exception_poison_pill()
        self.assertEqual(comms.get_exception(), (POISON_PILL, POISON_PILL))
        comms.task_done_exception()

        # Should be joinable. When using keep_alive, it should still be open
        comms.join_exception_queue(keep_alive=True)
        comms.add_exception(TypeError, 'TypeError')
        comms.get_exception()
        comms.task_done_exception()
        comms.join_exception_queue(keep_alive=False)

        # Depending on Python version it either throws AssertionError or ValueError
        with self.assertRaises((AssertionError, ValueError)):
            comms.add_exception(TypeError, 'TypeError')

    def test_exception_thrown(self):
        """
        Test exception thrown related functions
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 2, False, False)

        self.assertFalse(comms.exception_thrown())
        comms.set_exception_thrown()
        self.assertTrue(comms.exception_thrown())
        comms._exception_thrown.clear()
        self.assertFalse(comms.exception_thrown())

    def test_kill_signal_received(self):
        """
        Test kill signal received related functions
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 2, False, False)

        self.assertFalse(comms.kill_signal_received())
        comms.set_kill_signal_received()
        self.assertTrue(comms.kill_signal_received())
        comms._kill_signal_received.clear()
        self.assertFalse(comms.kill_signal_received())

    def test_worker_poison_pill(self):
        """
        Test that a poison pill is inserted for every worker
        """
        for n_jobs in [1, 2, 4]:
            with self.subTest(n_jobs=n_jobs):
                comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], n_jobs, False, False)
                comms.init_comms(False, False)
                comms.insert_poison_pill()
                for worker_id in range(n_jobs):
                    self.assertEqual(comms.get_task(worker_id), POISON_PILL)
                    comms.task_done(worker_id)
                comms.join_task_queues()

    def test_worker_non_lethal_poison_pill(self):
        """
        Test that a non-lethal poison pill is inserted for every worker
        """
        # Shouldn't be the same thing
        self.assertNotEqual(POISON_PILL, NON_LETHAL_POISON_PILL)

        for n_jobs in [1, 2, 4]:
            with self.subTest(n_jobs=n_jobs):
                comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], n_jobs, False, False)
                comms.init_comms(False, False)
                comms.insert_non_lethal_poison_pill()
                for worker_id in range(n_jobs):
                    self.assertEqual(comms.get_task(worker_id), NON_LETHAL_POISON_PILL)
                    comms.task_done(worker_id)
                comms.join_task_queues()

    def test_worker_restart(self):
        """
        Test worker restart related functions
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 5, False, False)
        comms.init_comms(False, False)

        # No restarts yet
        self.assertListEqual(list(comms.get_worker_restarts()), [])

        # Signal some restarts
        comms.signal_worker_restart(0)
        comms.signal_worker_restart(2)
        comms.signal_worker_restart(3)

        # Restarts available
        self.assertListEqual(list(comms.get_worker_restarts()), [0, 2, 3])

        # Reset some
        comms.reset_worker_restart(0)
        comms.reset_worker_restart(3)

        # Restarts available
        self.assertListEqual(list(comms.get_worker_restarts()), [2])

        # Reset last one
        comms.reset_worker_restart(2)
        self.assertListEqual(list(comms.get_worker_restarts()), [])

    def test_worker_alive(self):
        """
        Test worker alive related functions
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 5, False, False)
        comms.init_comms(False, False)

        # Signal some workers are alive
        comms.set_worker_alive(0)
        comms.set_worker_alive(1)
        comms.set_worker_dead(1)
        comms.set_worker_alive(2)
        comms.set_worker_alive(3)

        # Check alive status
        self.assertListEqual([comms.is_worker_alive(worker_id) for worker_id in range(5)],
                             [True, False, True, True, False])

        # Reset some
        comms.set_worker_dead(0)
        comms.set_worker_dead(3)

        # Check alive status
        self.assertListEqual([comms.is_worker_alive(worker_id) for worker_id in range(5)],
                             [False, False, True, False, False])

        # We test wait by simply checking the call count
        for worker_id in range(5):
            with patch.object(comms._workers_dead[worker_id], 'wait') as p:
                comms.wait_for_dead_worker(worker_id)
                self.assertEqual(p.call_count, 1)

    def test_drain_queues_terminate_worker(self):
        """
        get_results should be called once, get_exit_results should be called when exit function is defined
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 5, False, False)
        dont_wait_event = threading.Event()
        dont_wait_event.set()

        comms.init_comms(has_worker_exit=False, has_progress_bar=False)
        with self.subTest(has_worker_exit=False, has_progress_bar=False), \
                patch.object(comms, 'get_results', side_effect=comms.get_results) as p1, \
                patch.object(comms, 'get_exit_results') as p2:
            comms.drain_queues_terminate_worker(0, dont_wait_event)
            self.assertEqual(p1.call_count, 1)
            self.assertEqual(p2.call_count, 0)
            self.assertTrue(dont_wait_event.is_set())

        comms.init_comms(has_worker_exit=True, has_progress_bar=False)
        for worker_id in range(5):
            with self.subTest(has_worker_exit=True, has_progress_bar=False, worker_id=worker_id), \
                    patch.object(comms, 'get_results', side_effect=comms.get_results) as p1, \
                    patch.object(comms, 'get_exit_results', side_effect=comms.get_exit_results) as p2:
                comms.drain_queues_terminate_worker(worker_id, dont_wait_event)
                self.assertEqual(p1.call_count, 1)
                self.assertEqual(p2.call_count, 1)
                self.assertEqual(p2.call_args_list[0][0][0], worker_id)
                self.assertTrue(dont_wait_event.is_set())

        comms.init_comms(has_worker_exit=False, has_progress_bar=True)
        with self.subTest(has_worker_exit=False, has_progress_bar=True), \
                patch.object(comms, 'get_results', side_effect=comms.get_results) as p1, \
                patch.object(comms, 'get_exit_results') as p2, \
                patch.object(comms._task_completed_queue, 'get', side_effect=comms._task_completed_queue.get) as p3:
            comms.drain_queues_terminate_worker(0, dont_wait_event)
            self.assertEqual(p1.call_count, 1)
            self.assertEqual(p2.call_count, 0)
            self.assertEqual(p3.call_count, 1)
            self.assertTrue(dont_wait_event.is_set())

        comms.init_comms(has_worker_exit=True, has_progress_bar=True)
        for worker_id in range(5):
            with self.subTest(has_worker_exit=True, has_progress_bar=True, worker_id=worker_id), \
                    patch.object(comms, 'get_results', side_effect=comms.get_results) as p1, \
                    patch.object(comms, 'get_exit_results', side_effect=comms.get_exit_results) as p2, \
                    patch.object(comms._task_completed_queue, 'get', side_effect=comms._task_completed_queue.get) as p3:
                comms.drain_queues_terminate_worker(worker_id, dont_wait_event)
                self.assertEqual(p1.call_count, 1)
                self.assertEqual(p2.call_count, 1)
                self.assertEqual(p2.call_args_list[0][0][0], worker_id)
                self.assertEqual(p3.call_count, 1)
                self.assertTrue(dont_wait_event.is_set())

    def test_drain_queues(self):
        """
        _drain_and_join_queue should be called for every queue that matters. There are as many tasks queues as workers,
        1 results queue, if a worker_exit function has been provided as many exit results queues as workers, and if a
        progress bar has been enabled one queu for the progress bar.
        """
        for n_jobs, has_worker_exit, has_progress_bar in product([1, 2, 4], [False, True], [False, True]):
            comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], n_jobs, False, False)
            comms.init_comms(has_worker_exit=has_worker_exit, has_progress_bar=has_progress_bar)

            with self.subTest(n_jobs=n_jobs, has_worker_exit=has_worker_exit, has_progress_bar=has_progress_bar), \
                    patch.object(comms, 'drain_and_join_queue') as p:
                comms.drain_queues()
                self.assertEqual(p.call_count,
                                 n_jobs + 1 + (n_jobs if has_worker_exit else 0) + (1 if has_progress_bar else 0))

    def test__drain_and_join_queue(self):
        """
        Test draining queues
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 2, False, False)

        # Create a custom queue with some data
        q = mp.JoinableQueue()
        q.put(1)
        q.put('hello')
        q.put('world')

        # Drain queue. It should now be closed
        comms._drain_and_join_queue(q)

        # Depending on Python version it either throws OSError or ValueError
        with self.assertRaises((OSError, ValueError)):
            q.get(block=False)



class StopWorker(Exception):
    """ Exception used to kill workers from the main process """
    pass


class CannotPickleExceptionError(Exception):
    """ Exception used when Pickle has trouble pickling the actual Exception """
    pass



import ctypes
import math
import multiprocessing.context
import multiprocessing.managers
from datetime import datetime
from functools import partial
from typing import Dict, Optional, List, Tuple

from mpire.context import RUNNING_WINDOWS
from mpire.signal import ignore_keyboard_interrupt
from mpire.utils import format_seconds


class WorkerInsights:

    """
    Worker insights class for profiling the worker start up time, waiting time and working time. When worker init and
    exit functions are provided it will time those as well.
    """

    def __init__(self, ctx: multiprocessing.context.BaseContext, n_jobs: int) -> None:
        """
        Parameter class for worker insights.

        :param ctx: Multiprocessing context
        :param n_jobs: Number of workers
        """
        self.ctx = ctx
        self.n_jobs = n_jobs

        # Whether insights have been enabled or not
        self.insights_enabled = False

        # Multiprocessing Manager for storing the max_task_args in. A lock is needed to ensure no race conditions occur
        self.insights_manager = None
        self.insights_manager_lock = None

        # `datetime` object indicating at what time the Worker instance was created and started
        self.worker_start_up_time = None

        # Array object which holds the total number of seconds the workers take to start up
        self.worker_init_time = None

        # Array object which holds the total number of completed tasks per worker
        self.worker_n_completed_tasks = None

        # Array object which holds the total number of seconds the workers have been idle
        self.worker_waiting_time = None

        # Array object which holds the total number of seconds the workers are executing the task function
        self.worker_working_time = None

        # Array object which holds the total number of seconds the workers take to run the exit function
        self.worker_exit_time = None

        # Array object which holds the top 5 max task durations in seconds per worker
        self.max_task_duration = None

        # Manager.List object which holds the top 5 task arguments (string) for the longest task per worker
        self.max_task_args = None

    def reset_insights(self, enable_insights: bool) -> None:
        """
        Resets the insights containers

        :param enable_insights: Whether to enable worker insights
        """
        if enable_insights:
            # When we're on Windows, we don't use a Manager as it's giving authentication errors. The max_task_args
            # information is therefore not available on Windows systems. This needs to be fixed at some point in time
            if not RUNNING_WINDOWS:
                # We need to ignore the KeyboardInterrupt signal for the manager to avoid BrokenPipeErrors
                self.insights_manager = multiprocessing.managers.SyncManager(ctx=self.ctx)
                self.insights_manager.start(ignore_keyboard_interrupt)
            self.insights_manager_lock = self.ctx.Lock()
            self.worker_start_up_time = self.ctx.Array(ctypes.c_double, self.n_jobs, lock=False)
            self.worker_init_time = self.ctx.Array(ctypes.c_double, self.n_jobs, lock=False)
            self.worker_n_completed_tasks = self.ctx.Array(ctypes.c_int, self.n_jobs, lock=False)
            self.worker_waiting_time = self.ctx.Array(ctypes.c_double, self.n_jobs, lock=False)
            self.worker_working_time = self.ctx.Array(ctypes.c_double, self.n_jobs, lock=False)
            self.worker_exit_time = self.ctx.Array(ctypes.c_double, self.n_jobs, lock=False)
            self.max_task_duration = self.ctx.Array(ctypes.c_double, self.n_jobs * 5, lock=False)
            if RUNNING_WINDOWS:
                # Doesn't actually do anything, but reduces the amount of if/else statements in other code parts
                self.max_task_args = [""] * self.n_jobs * 5
            else:
                self.max_task_args = self.insights_manager.list([""] * self.n_jobs * 5)
        else:
            self.insights_manager = None
            self.insights_manager_lock = None
            self.worker_start_up_time = None
            self.worker_init_time = None
            self.worker_n_completed_tasks = None
            self.worker_waiting_time = None
            self.worker_working_time = None
            self.worker_exit_time = None
            self.max_task_duration = None
            self.max_task_args = None

        self.insights_enabled = enable_insights

    def get_max_task_duration_list(self, worker_id: int) -> Optional[List[Tuple[float, str]]]:
        """
        Initialize insights for a specific worker

        :param worker_id: worker ID
        """
        if self.insights_enabled:
            # Local worker insights container that holds (task duration, task args) tuples, sorted for heapq. We use a
            # local container for each worker as to not put too big of a burden on interprocess communication
            with self.insights_manager_lock:
                return (list(zip(self.max_task_duration[worker_id * 5:(worker_id + 1) * 5],
                                 self.max_task_args[worker_id * 5:(worker_id + 1) * 5]))
                        if self.max_task_duration is not None else None)

    def update_start_up_time(self, worker_id: int, start_time: datetime) -> None:
        """
        Update start up time

        :param worker_id: Worker ID
        :param start_time: datetime
        """
        if self.insights_enabled:
            self.worker_start_up_time[worker_id] = (datetime.now() - start_time).total_seconds()

    def update_n_completed_tasks(self, worker_id: int) -> None:
        """
        Increment the number of completed tasks for this worker

        :param worker_id: Worker ID
        """
        if self.insights_enabled:
            self.worker_n_completed_tasks[worker_id] += 1

    def update_task_insights(self, worker_id: int, max_task_duration_last_updated: datetime,
                             max_task_duration_list: Optional[List[Tuple[float, str]]],
                             force_update: bool = False) -> datetime:
        """
        Update synced containers with new top 5 max task duration + args. Updates every 2 seconds.

        :param worker_id: Worker ID
        :param max_task_duration_last_updated: Last updated datetime
        :param max_task_duration_list: Local worker insights container that holds (task duration, task args) tuples,
            sorted for heapq
        :param force_update: Whether to force the update
        :return: Last updated datetime
        """
        now = datetime.now()
        if self.insights_enabled and (force_update or (now - max_task_duration_last_updated).total_seconds() > 2):
            task_durations, task_args = zip(*max_task_duration_list)
            self.max_task_duration[worker_id * 5:(worker_id + 1) * 5] = task_durations
            with self.insights_manager_lock:
                self.max_task_args[worker_id * 5:(worker_id + 1) * 5] = task_args
            max_task_duration_last_updated = now

        return max_task_duration_last_updated

    def get_insights(self) -> Dict:
        """
        Creates insights from the raw insight data

        :return: dictionary containing worker insights
        """
        def argsort(seq):
            """
            argsort, as to not be dependent on numpy, by
            https://stackoverflow.com/questions/3382352/equivalent-of-numpy-argsort-in-basic-python/3382369#3382369
            """
            return sorted(range(len(seq)), key=seq.__getitem__)

        def mean_std(seq):
            """
            Calculates mean and standard deviation, as to not be dependent on numpy
            """
            _mean = sum(seq) / len(seq)
            _var = sum(pow(x - _mean, 2) for x in seq) / len(seq)
            _std = math.sqrt(_var)
            return _mean, _std

        if not self.insights_enabled:
            return {}

        format_seconds_func = partial(format_seconds, with_milliseconds=True)

        # Determine max 5 tasks based on duration, exclude zero values and args that haven't been synced yet (empty str)
        sorted_idx = argsort(self.max_task_duration)[-5:][::-1]
        top_5_max_task_durations, top_5_max_task_args = [], []
        for idx in sorted_idx:
            if self.max_task_duration[idx] == 0:
                break
            if self.max_task_args[idx] == "" and not RUNNING_WINDOWS:
                continue
            top_5_max_task_durations.append(format_seconds_func(self.max_task_duration[idx]))
            top_5_max_task_args.append("" if RUNNING_WINDOWS else self.max_task_args[idx])

        # Populate
        total_start_up_time = sum(self.worker_start_up_time)
        total_init_time = sum(self.worker_init_time)
        total_waiting_time = sum(self.worker_waiting_time)
        total_working_time = sum(self.worker_working_time)
        total_exit_time = sum(self.worker_exit_time)
        total_time = total_start_up_time + total_init_time + total_waiting_time + total_working_time + total_exit_time
        insights = dict(n_completed_tasks=list(self.worker_n_completed_tasks),
                        start_up_time=list(map(format_seconds_func, self.worker_start_up_time)),
                        init_time=list(map(format_seconds_func, self.worker_init_time)),
                        waiting_time=list(map(format_seconds_func, self.worker_waiting_time)),
                        working_time=list(map(format_seconds_func, self.worker_working_time)),
                        exit_time=list(map(format_seconds_func, self.worker_exit_time)),
                        total_start_up_time=format_seconds_func(total_start_up_time),
                        total_init_time=format_seconds_func(total_init_time),
                        total_waiting_time=format_seconds_func(total_waiting_time),
                        total_working_time=format_seconds_func(total_working_time),
                        total_exit_time=format_seconds_func(total_exit_time),
                        top_5_max_task_durations=top_5_max_task_durations,
                        top_5_max_task_args=top_5_max_task_args)

        insights['total_time'] = format_seconds_func(total_time)

        # Calculate ratio, mean and standard deviation of different parts of the worker lifespan
        for part, total in (('start_up', total_start_up_time),
                            ('init', total_init_time),
                            ('waiting', total_waiting_time),
                            ('working', total_working_time),
                            ('exit', total_exit_time)):
            mean, std = mean_std(getattr(self, f'worker_{part}_time'))
            insights[f'{part}_ratio'] = total / (total_time + 1e-8)
            insights[f'{part}_time_mean'] = format_seconds_func(mean)
            insights[f'{part}_time_std'] = format_seconds_func(std)

        return insights

    def get_insights_string(self) -> str:
        """
        Formats the worker insights_str and returns a string

        :return: worker insights_str string
        """
        if not self.insights_enabled:
            return "No profiling stats available. Try to run a function first with insights enabled ..."

        insights = self.get_insights()
        insights_str = ["WorkerPool insights",
                        "-------------------",
                        f"Total number of tasks completed: {sum(insights['n_completed_tasks'])}"]

        # Format string for parts of the worker lifespan
        for part in ('start_up', 'init', 'waiting', 'working', 'exit'):
            insights_str.append(f"Total {part.replace('_', ' ')} time: {insights[f'total_{part}_time']}s ("
                                f"mean: {insights[f'{part}_time_mean']}, std: {insights[f'{part}_time_std']}, "
                                f"ratio: {insights[f'{part}_ratio'] * 100.:.2f}%)")

        # Add warning when working ratio is below 80%
        if insights['working_ratio'] < 0.8:
            insights_str.extend(["",
                                 "Efficiency warning: working ratio is < 80%!"])

        # Add stats per worker
        insights_str.extend(["",
                             "Stats per worker",
                             "----------------"])
        for worker_id in range(self.n_jobs):
            worker_str = [f"Worker {worker_id}",
                          f"Tasks completed: {insights['n_completed_tasks'][worker_id]}"]
            for part in ('start_up', 'init', 'waiting', 'working', 'exit'):
                worker_str.append(f"{part.replace('_', ' ')}: {insights[f'{part}_time'][worker_id]}s")
            insights_str.append(' - '.join(worker_str))

        # Add task stats
        insights_str.extend(["",
                             "Top 5 longest tasks",
                             "-------------------"])
        max_task_duration_args_separator = "" if RUNNING_WINDOWS else " - "
        for task_idx, (duration, args) in enumerate(zip(insights['top_5_max_task_durations'],
                                                        insights['top_5_max_task_args']), start=1):
            insights_str.append(f"{task_idx}. Time: {duration}{max_task_duration_args_separator}{args}")

        return "\n".join(insights_str)



import logging
import multiprocessing as mp
import multiprocessing.context
import sys
from datetime import datetime, timedelta
from typing import Any, Callable, Dict, Optional, Type

from tqdm.auto import tqdm

from mpire.comms import WorkerComms, POISON_PILL
from mpire.dashboard.connection_utils import (DashboardConnectionDetails, get_dashboard_connection_details,
                                              set_dashboard_connection)
from mpire.insights import WorkerInsights
from mpire.signal import DisableKeyboardInterruptSignal, ignore_keyboard_interrupt
from mpire.tqdm_utils import TqdmConnectionDetails, TqdmManager
from mpire.utils import format_seconds

# If a user has not installed the dashboard dependencies than the imports below will fail
try:
    from mpire.dashboard.dashboard import DASHBOARD_STARTED_EVENT
    from mpire.dashboard.utils import get_function_details
    from mpire.dashboard.manager import get_manager_client_dicts
except ImportError:
    DASHBOARD_STARTED_EVENT = None

    def get_function_details(_):
        pass

    def get_manager_client_dicts():
        raise NotImplementedError

logger = logging.getLogger(__name__)

DATETIME_FORMAT = "%Y-%m-%d, %H:%M:%S"


class ProgressBarHandler:

    def __init__(self, ctx: multiprocessing.context.BaseContext, using_threading: bool, func: Callable, n_jobs: int,
                 show_progress_bar: bool, progress_bar_total: int, progress_bar_position: int,
                 worker_comms: WorkerComms, worker_insights: WorkerInsights) -> None:
        """
        :param ctx: Multiprocessing context
        :param using_threading: Whether threading is used as backend
        :param func: Function passed on to a WorkerPool map function
        :param n_jobs: Number of workers that are used
        :param show_progress_bar: When ``True`` will display a progress bar
        :param progress_bar_total: Total number of tasks that will be processed
        :param progress_bar_position: Denotes the position (line nr) of the progress bar. This is useful wel using
            multiple progress bars at the same time
        :param worker_comms: Worker communication objects (queues, locks, events, ...)
        :param worker_insights: WorkerInsights object which stores the worker insights
        """
        # When the threading backend is used we switch to a multiprocessing context, because the progress bar handler
        # needs to be a process, not a thread. This is because when using threading, the progress bar updates will
        # interfere too much with the main process.
        self.ctx = mp if using_threading else ctx
        self.show_progress_bar = show_progress_bar
        self.progress_bar_total = progress_bar_total
        self.progress_bar_position = progress_bar_position
        self.worker_comms = worker_comms
        self.worker_insights = worker_insights
        if show_progress_bar and DASHBOARD_STARTED_EVENT is not None:
            self.function_details = get_function_details(func)
            self.function_details['n_jobs'] = n_jobs
        else:
            self.function_details = None

        self.process = None
        self.process_started = self.ctx.Event()
        self.progress_bar_id = None
        self.dashboard_dict = None
        self.dashboard_details_dict = None
        self.start_t = None

    def __enter__(self) -> 'ProgressBarHandler':
        """
        Enables the use of the ``with`` statement. Starts a new progress handler process if a progress bar should be
        shown

        :return: self
        """
        if self.show_progress_bar:

            # Disable the interrupt signal. We let the process die gracefully
            with DisableKeyboardInterruptSignal():

                # We start a new process because updating the progress bar in a thread can slow down processing
                # of results and can fail to show real-time updates
                logger.debug("Starting progress bar handler")
                self.process = self.ctx.Process(target=self._progress_bar_handler,
                                                args=(TqdmManager.get_connection_details(),
                                                      get_dashboard_connection_details(),))
                self.process.start()
                self.process_started.wait()

        return self

    def __exit__(self, exc_type: Type, *_) -> None:
        """
        Enables the use of the ``with`` statement. Terminates the progress handler process if there is one
        """
        if self.show_progress_bar and self.process.is_alive():

            # If this exit is called with an exception, then we assume an external kill signal was received (this is,
            # for example, necessary in nested pools when an error occurs)
            if exc_type is not None:
                self.worker_comms.set_kill_signal_received()

            # Insert poison pill and close the handling process
            if not self.worker_comms.exception_thrown():
                logger.debug("Adding poison pill to progress bar")
                self.worker_comms.add_progress_bar_poison_pill()
            logger.debug("Joining progress bar handler")
            self.process.join()
            logger.debug("Progress bar handler joined")

    def _progress_bar_handler(self, tqdm_connection_details: TqdmConnectionDetails,
                              dashboard_connection_details: DashboardConnectionDetails) -> None:
        """
        Keeps track of the progress made by the workers and updates the progress bar accordingly

        :param tqdm_connection_details: Tqdm manager host, and whether the manager is started/connected
        :param dashboard_connection_details: Dashboard manager host, port_nr and whether a dashboard is
            started/connected
        """
        ignore_keyboard_interrupt()  # For Windows compatibility

        logger.debug("Progress bar handler started")
        self.process_started.set()

        # Set tqdm and dashboard connection details. This is needed for nested pools and in the case forkserver or
        # spawn is used as start method
        TqdmManager.set_connection_details(tqdm_connection_details)
        set_dashboard_connection(dashboard_connection_details)

        # Connect to the tqdm manager
        tqdm_manager = TqdmManager()
        tqdm_lock, tqdm_position_register = tqdm_manager.get_lock_and_position_register()
        tqdm.set_lock(tqdm_lock)
        main_progress_bar = tqdm_position_register.register_progress_bar_position(self.progress_bar_position)

        # In case we're running tqdm in a notebook we need to apply a dirty hack to get progress bars working.
        # Solution adapted from https://github.com/tqdm/tqdm/issues/485#issuecomment-473338308
        in_notebook = 'IPython' in sys.modules and 'IPKernelApp' in sys.modules['IPython'].get_ipython().config
        if in_notebook:
            print(' ', end='', flush=True)

        # Create progress bar and register the start time
        progress_bar = tqdm(total=self.progress_bar_total, position=self.progress_bar_position, dynamic_ncols=True,
                            leave=True)
        self.start_t = datetime.fromtimestamp(progress_bar.start_t)

        # Register progress bar to dashboard in case a dashboard is started
        self._register_progress_bar(progress_bar)

        while True:
            # Wait for a job to finish
            tasks_completed, from_queue = self.worker_comms.get_tasks_completed_progress_bar()

            # If we received a poison pill, we should quit right away. We do force a final refresh of the progress bar
            # to show the latest status
            if tasks_completed is POISON_PILL:
                logger.debug("Terminating progress bar handler")

                # Check if we got a poison pill because there was an error. If so, we obtain the exception information
                # and send it to the dashboard, if available.)
                if self.worker_comms.exception_thrown() or self.worker_comms.kill_signal_received():
                    progress_bar.set_description('Exception occurred, terminating ... ')
                    if self.worker_comms.exception_thrown():
                        _, traceback_str = self.worker_comms.get_exception()
                        self._send_update(progress_bar, failed=True, traceback_str=traceback_str)
                        self.worker_comms.task_done_exception()
                    elif self.worker_comms.kill_signal_received():
                        self._send_update(progress_bar, failed=True, traceback_str='Kill signal received')

                # Final update of the progress bar. When we're not in a notebook and this is the main progress bar, we
                # add as many newlines as the highest progress bar position, such that new output is added after the
                # progress bars.
                progress_bar.refresh()
                if in_notebook:
                    progress_bar.close()
                else:
                    progress_bar.disable = True
                    if main_progress_bar:
                        progress_bar.fp.write('\n' * (tqdm_position_register.get_highest_progress_bar_position() + 1))
                if from_queue:
                    self.worker_comms.task_done_progress_bar()
                break

            # Update progress bar
            progress_bar.update(tasks_completed)
            self.worker_comms.task_done_progress_bar()

            # Force a refresh when we're at 100%
            if progress_bar.n == progress_bar.total:
                if in_notebook:
                    progress_bar.close()
                progress_bar.refresh()
                self.worker_comms.set_progress_bar_complete()
                self.worker_comms.wait_until_progress_bar_is_complete()
                self._send_update(progress_bar)

            # Send update to dashboard in case a dashboard is started, but only when tqdm updated its view as well. This
            # will make the dashboard a lot more responsive
            if progress_bar.n == progress_bar.last_print_n:
                self._send_update(progress_bar)

        logger.debug("Progress bar handler done")

    def _register_progress_bar(self, progress_bar: tqdm) -> None:
        """
        Register this progress bar to the dashboard

        :param progress_bar: tqdm progress bar instance
        """
        if self.progress_bar_id is None and DASHBOARD_STARTED_EVENT is not None and DASHBOARD_STARTED_EVENT.is_set():

            # Connect to manager server
            self.dashboard_dict, self.dashboard_details_dict, dashboard_tqdm_lock = get_manager_client_dicts()

            # Register new progress bar
            logger.debug("Registering new progress bar to the dashboard server")
            dashboard_tqdm_lock.acquire()
            self.progress_bar_id = len(self.dashboard_dict.keys()) + 1
            self.dashboard_details_dict.update([(self.progress_bar_id, self.function_details)])
            self._send_update(progress_bar)
            dashboard_tqdm_lock.release()

    def _send_update(self, progress_bar: tqdm, failed: bool = False, traceback_str: Optional[str] = None) -> None:
        """
        Adds a progress bar update to the shared dict so the dashboard process can use it, only when a dashboard has
        started

        :param progress_bar: tqdm progress bar instance
        :param failed: Whether or not the operation failed or not
        :param traceback_str: Traceback string, if an exception was raised
        """
        if self.progress_bar_id is not None:
            self.dashboard_dict.update([(self.progress_bar_id,
                                         self._get_progress_bar_update_dict(progress_bar, failed, traceback_str))])

    def _get_progress_bar_update_dict(self, progress_bar: tqdm, failed: bool,
                                      traceback_str: Optional[str] = None) -> Dict[str, Any]:
        """
        Obtain update dictionary with all the information needed for displaying on the dashboard

        :param progress_bar: tqdm progress bar instance
        :param failed: Whether or not the operation failed or not
        :param traceback_str: Traceback string, if an exception was raised
        :return: Update dictionary
        """
        # Save some variables first so we can use them consistently with the same value
        details = progress_bar.format_dict
        n = details["n"]
        total = details["total"]
        now = datetime.now()
        remaining_time = (total - n) / details["rate"] if details["rate"] else None

        return {"id": self.progress_bar_id,
                "success": not failed,
                "n": n,
                "total": total,
                "percentage": n / total,
                "duration": str(now - self.start_t).rsplit('.', 1)[0],
                "remaining": format_seconds(remaining_time, False),
                "started_raw": self.start_t,
                "started": self.start_t.strftime(DATETIME_FORMAT),
                "finished_raw": now + timedelta(seconds=remaining_time) if remaining_time is not None else None,
                "finished": ((now + timedelta(seconds=remaining_time)).strftime(DATETIME_FORMAT)
                             if remaining_time is not None else ''),
                "traceback": traceback_str.strip() if traceback_str is not None else None,
                "insights": self.worker_insights.get_insights()}
